#include "stdafx.h"
#include "avifile.h"
#include "images.h"




int FrameRate=30;






CAviWrite::CAviWrite()
{
	aopts[0]=&opts;
	pfile=NULL;
	ps=NULL;
	psCompressed=NULL;
	index=0;
	AVIFileInit();
}

CAviWrite::~CAviWrite()
{
	if (ps)
		AVIStreamClose(ps);

	if (psCompressed)
		AVIStreamClose(psCompressed);

	if (pfile)
		AVIFileClose(pfile);

	AVIFileExit();
}

BOOLEAN CAviWrite::Create(PCHAR file)
{
	// Creates a new AVI file
	if (AVIFileOpen(&pfile,
		       file,
		       OF_WRITE | OF_CREATE,
		       NULL)!=AVIERR_OK)return FALSE;

	return TRUE;
}

BOOLEAN CAviWrite::InitVideoStream(PCHAR file)
{
	// Creates a video stream for frames
	HBITMAP bmp=LoadPicture(file);
	if (bmp==0)return FALSE;
	BITMAPINFOHEADER *bih;
	HANDLE dib=MakeDib(bmp,24);
	DeleteObject(bmp);

	bih=(LPBITMAPINFOHEADER)GlobalLock(dib);

	ZeroMemory(&strhdr,sizeof(strhdr));
	strhdr.fccType                = streamtypeVIDEO;
	strhdr.fccHandler             = 0;
	strhdr.dwScale                = 1;
	strhdr.dwRate                 = FrameRate;		    
	strhdr.dwSuggestedBufferSize  = bih->biSizeImage;
	SetRect(&strhdr.rcFrame, 0, 0,		  
	    (int) bih->biWidth,
	    (int) bih->biHeight);

	width=bih->biWidth;
	height=bih->biHeight;

	// And create the stream;
	if (AVIFileCreateStream(pfile,		    // file pointer
			         &ps,		    // returned stream pointer
			         &strhdr)!=AVIERR_OK)
	{
		GlobalUnlock(dib);
		GlobalFree(dib);
		return FALSE;
	}

	ZeroMemory(&opts,sizeof(opts));

	if (!AVISaveOptions(NULL, 0, 1, &ps, (LPAVICOMPRESSOPTIONS FAR *) &aopts))
	{
		GlobalUnlock(dib);
		GlobalFree(dib);
		return FALSE;
	}

	if (AVIMakeCompressedStream(&psCompressed, ps, &opts, NULL)!=AVIERR_OK)
	{
		GlobalUnlock(dib);
		GlobalFree(dib);
		return FALSE;
	}

	if (AVIStreamSetFormat(psCompressed, 0,
			       bih,	    // stream format
			       bih->biSize +   // format size
			       bih->biClrUsed * sizeof(RGBQUAD))!=AVIERR_OK)
	{
		GlobalUnlock(dib);
		GlobalFree(dib);
		return FALSE;
	}

	GlobalUnlock(dib);
	GlobalFree(dib);

	return TRUE;
}

BOOLEAN CAviWrite::Add(PCHAR file)
{
	// Adds a simple file to the AVI
	HBITMAP bitmap=LoadPicture(file);
	// If it's no picture, add as audio
	if (bitmap==0)return AddAudio(file);

	if (ps==NULL)
	{	// Create video stream, if not already done
		if (!InitVideoStream(file))return FALSE;
	}

	// Add picture to video stream
	BITMAP b;
	GetObject(bitmap,sizeof(b),&b);

	if ((b.bmWidth!=width)||(b.bmHeight!=height))
	{	// Stretch it
		HDC deskdc=GetDC(0);
		HDC dc1=CreateCompatibleDC(deskdc),dc2=CreateCompatibleDC(deskdc);
		HBITMAP bmp=CreateCompatibleBitmap(deskdc,width,height);
		ReleaseDC(0,deskdc);

		SelectObject(dc1,bmp);
		SelectObject(dc2,bitmap);

		SetStretchBltMode(dc1,HALFTONE);

		StretchBlt(dc1,0,0,width,height,dc2,0,0,b.bmWidth,b.bmHeight,SRCCOPY);

		RestoreDC(dc1,-1);
		RestoreDC(dc2,-1);

		DeleteDC(dc2);
		DeleteDC(dc1);
		DeleteObject(bitmap);

		bitmap=bmp;
	}

	HANDLE dib=MakeDib(bitmap,24);
	if (dib==0)return FALSE;
	DeleteObject(bitmap);
	LPBITMAPINFOHEADER bi=(LPBITMAPINFOHEADER)GlobalLock(dib);

	HRESULT hr;
	
	hr = AVIStreamWrite(psCompressed,	// stream pointer
			index,				// time of this frame
			1,				// number to write
			(LPBYTE) bi +		// pointer to data
				bi->biSize +
				bi->biClrUsed * sizeof(RGBQUAD),
				bi->biSizeImage,	// size of this frame
			AVIIF_KEYFRAME,			 // flags....
			NULL,
			NULL);

	if (GlobalUnlock(dib)==TRUE)return FALSE;
	int e=GetLastError();
	if (GlobalFree(dib)!=0)return FALSE;
	index++;
	if (hr != AVIERR_OK)return FALSE;

	return TRUE;
}

BOOLEAN CAviWrite::AddAudio(PCHAR filename)
{
	// Loads and saves an wave form audio file into an audio stream
	HANDLE file=CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,0);
	if (file==INVALID_HANDLE_VALUE)return FALSE;
	if (file==0)return FALSE;

	long l;
	CHAR b[5];
	b[4]='\0';
	DWORD read;
	LPVOID format=NULL,data=NULL;
	long formatsize,datasize;
	PAVISTREAM audio=NULL;

	ReadFile(file,&b[0],4,&read,NULL);
	if (strcmp(b,"RIFF")!=0)goto error;
	
	ReadFile(file,&l,sizeof(l),&read,NULL);

	ReadFile(file,&b[0],4,&read,NULL);
	if (strcmp(b,"WAVE")!=0)goto error;

	ReadFile(file,&b[0],4,&read,NULL);

	// read format 
	ReadFile(file,&formatsize,sizeof(formatsize),&read,NULL);
	format=malloc(formatsize);

	ReadFile(file,format,formatsize,&read,NULL);

	// read data 
	ReadFile(file,&b[0],4,&read,NULL);
	if (strcmp(b,"data")!=0)goto error;

	ReadFile(file,&datasize,sizeof(datasize),&read,NULL);
	data=malloc(datasize);

	ReadFile(file,data,datasize,&read,NULL);
	if (DWORD(datasize)!=read)return FALSE;

	// All read, now create audio stream and write 

	AVISTREAMINFO strhdr;

	ZeroMemory(&strhdr,sizeof(strhdr));
	strhdr.fccType                = streamtypeAUDIO;
	strhdr.fccHandler             = 0;
	strhdr.dwScale=strhdr.dwSampleSize= ((WAVEFORMAT*)(format))->nBlockAlign;
	strhdr.dwRate                 = ((WAVEFORMAT*)(format))->nSamplesPerSec*strhdr.dwScale;		    
	strhdr.dwSuggestedBufferSize  = 0;
	
	if (AVIFileCreateStream(pfile,&audio,		   
			         &strhdr)!=AVIERR_OK)goto error;

	AVIStreamSetFormat(audio,0,format,formatsize);

	long samples,bytes;
	AVIStreamWrite(audio,0,datasize/strhdr.dwSampleSize,data,datasize,0,&samples,&bytes);

	// Ready, now clean up

	AVIStreamRelease(audio);
	free(format);
	free(data);
	CloseHandle(file);
	return TRUE;
error:
	if (format!=NULL)free(format);
	if (data!=NULL)free(data);
	CloseHandle(file);
	return FALSE;
}

