#include "stdafx.h"
#include "avifile.h"
#include "math.h"
#include "images.h"



CAviFile::CAviFile()
{
	ArcName=NULL;
	avi=NULL;
	stream=NULL;
	audio=NULL;
	frame=NULL;
	index=0;
	hdd=0;
	filesize=0;
	AVIFileInit();
}

CAviFile::~CAviFile()
{
	if (hdd)DrawDibClose(hdd);
	if (frame!=NULL)AVIStreamGetFrameClose(frame);
	if (audio)AVIStreamRelease(audio);
	if (stream)AVIStreamRelease(stream);
	if (avi)AVIFileRelease(avi);
	if (ArcName)free(ArcName);
	AVIFileExit();
}


BOOLEAN CAviFile::Load(PCHAR file)
{
	// Load an avi file and create streams
	ArcName=(PCHAR)malloc(strlen(file)+2);
	strcpy(ArcName,file);

	GetFileAttributesEx(file,GetFileExInfoStandard,&attr);

    HRESULT hr;

    hr = AVIFileOpen(&avi, file, 0, 0L);

    if (hr != 0)return FALSE;

	if (AVIFileGetStream(avi, &stream, streamtypeVIDEO, 0) != AVIERR_OK)stream=NULL; // no video stream
	if (AVIFileGetStream(avi, &audio, streamtypeAUDIO, 0) != AVIERR_OK)audio=NULL; // no audio stream


	if (audio)
	{
		AVIStreamInfo(audio,&audioinfo,sizeof(audioinfo));
		long datasize,formatsize;
		AVIStreamReadFormat(audio,0,NULL,&formatsize);
		AVIStreamRead(audio,0,audioinfo.dwLength,NULL,0,&datasize,NULL);

		audiosize=formatsize+datasize;
	}else audiosize=0;

	index=0;

	if (stream)
	{
		AVIStreamInfo(stream,&info,sizeof(info));
		frame=AVIStreamGetFrameOpen(stream,NULL);

		hdd=DrawDibOpen();

		{	// Determinate the file size of the frames
/*			HBITMAP bmp=GetBitmapFrame(0);
			HDC deskdc=GetDC(GetDesktopWindow());
			HDC dc=CreateCompatibleDC(deskdc);
	
			HBITMAP oldbmp=(HBITMAP)SelectObject(dc,bmp);
	

			BITMAPINFO  bmi;

			bmi.bmiHeader.biSize = 0x28;              // GDI need this to work
			bmi.bmiHeader.biBitCount = 0;             // don't get the color table
			GetDIBits(dc, bmp, 0, 0, (LPSTR)NULL, &bmi, DIB_RGB_COLORS);

			SelectObject(dc,oldbmp);
			DeleteObject(bmp);

			DeleteDC(dc);
			ReleaseDC(GetDesktopWindow(),deskdc);

			DWORD sizBMI=0;
			switch (bmi.bmiHeader.biBitCount) 
			{
				case 24:                                      // has color table
					sizBMI = sizeof(BITMAPINFOHEADER);
					break;
				case 16:
				case 32:
		            sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(DWORD)*3;
					break;
				default:
		            sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*(1<<bmi.bmiHeader.biBitCount);
					break;
			}
	
			filesize=bmi.bmiHeader.biSizeImage+sizBMI+sizeof(BITMAPFILEHEADER);*/

			LPBITMAPINFOHEADER pdata;
			pdata=GetDIBitmapFrame(0);
			if (pdata) {
				DWORD sizBMI=0;
				switch (pdata->biBitCount) 
				{
					case 24:                                      // has color table
						sizBMI = sizeof(BITMAPINFOHEADER);
						break;
					case 16:
					case 32:
						if (pdata->biCompression==3)
							sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(DWORD)*3;
						else
							sizBMI = sizeof(BITMAPINFOHEADER);
						break;
					default:
						sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*(1<<pdata->biBitCount);
						break;
				}
				filesize=pdata->biSizeImage+sizBMI+sizeof(BITMAPFILEHEADER);
			} else
				filesize=0;
		}
	}

	SYSTEMTIME time;
	FILETIME local;
	FileTimeToLocalFileTime(&attr.ftLastWriteTime,&local);
	FileTimeToSystemTime(&local,&time);
	filetime=FileTime(time.wYear,time.wMonth,time.wDay,time.wHour,time.wMinute,time.wSecond);

	return TRUE;
}

int CAviFile::ReadHeader(tHeaderData *HeaderData)
{
	// Returns the header of the avi file (the containing frames / audio)
	if (index>info.dwLength)return E_END_ARCHIVE;
	if ((index==info.dwLength)||(stream==NULL))
	{	// return AUDIO file, if available
		if (audio==NULL)return E_END_ARCHIVE;
		if ((stream==NULL)&&(index>=1))return E_END_ARCHIVE; // Audio already exported

		strcpy(HeaderData->ArcName,ArcName);
	
		strcpy(HeaderData->FileName,"Audio.wav");

		HeaderData->FileAttr=0x20;
		HeaderData->FileCRC=0;

		HeaderData->FileTime=filetime;


		HeaderData->UnpSize=audiosize+4+4+4+4+4+4+4;
		HeaderData->PackSize=audiosize;

		return 0;
	}

	strcpy(HeaderData->ArcName,ArcName);
	
	static CHAR format[20];
	int digits=(int)ceil(log10(info.dwLength+1));

	sprintf(&format[0],"Frame%%0.%dd.bmp",digits);

	sprintf(&HeaderData->FileName[0],format,index);
//	strcpy(HeaderData->FileName,&c[0]);

	HeaderData->FileAttr=0x20;
	HeaderData->FileCRC=0;

	HeaderData->FileTime=filetime;
		
	HeaderData->PackSize=(attr.nFileSizeLow-audiosize)/info.dwLength;
	HeaderData->UnpSize=filesize;

	return 0;
}

int CAviFile::ProcessFile(int Operation,char *DestPath,char *DestName)
{
	// Processes an extraction task
	if (Operation==PK_EXTRACT)
	{
		static CHAR path[MAX_PATH];
		if (DestPath==NULL)strcpy(&path[0],DestName);
		else {
			strcpy(&path[0],DestPath);
			strcat(&path[0],DestName);
		}

		if ((stream==NULL)||(index==info.dwLength))
		{	// Extract audio stream
			if (audio!=NULL)
			{
				return ExtractAudio(path);
			}
			return E_END_ARCHIVE;
		}

		// Extract frame 
		LPBITMAPINFOHEADER pdata;

		pdata=GetDIBitmapFrame(index);

		if (pdata) 
		{
			if (!SaveDIBitmap(pdata,&path[0]))
				return E_EWRITE;
		}else 
			return E_EREAD;

		if (ProcessDataProc(DestName,filesize)==0)
		{
			index++;
			return E_EABORTED;
		} 
	}

	index++;

	return 0;
}

void CAviFile::SetProcessDataProc(tProcessDataProc pProcessDataProc)
{
	// Set the callbackproc for processing the process dialog
	ProcessDataProc=pProcessDataProc;
}

int CAviFile::ExtractAudio(PCHAR filename)
{
	// Saves the audio stream to the specified file
	if (audio==NULL)return E_NO_FILES;
	HANDLE file=CreateFile(filename,GENERIC_WRITE,FILE_SHARE_WRITE,NULL,CREATE_ALWAYS,0,0);
	if (file==INVALID_HANDLE_VALUE)return E_ECREATE;

	LPWAVEFORMAT format;
	LPVOID data;
	long formatsize,datasize;
	{	// First extract format
		AVIStreamReadFormat(audio,0,NULL,&formatsize);
		format=(LPWAVEFORMAT)malloc(formatsize);
		AVIStreamReadFormat(audio,0,format,&formatsize);
		if (ProcessDataProc(filename,formatsize/2)==0)
		{
			free(format);
			CloseHandle(file);
			return E_EABORTED;
		}
	}
	{	// Extract wave data
		AVIStreamRead(audio,0,audioinfo.dwLength,NULL,0,&datasize,NULL);
		data=malloc(datasize);
		AVIStreamRead(audio,0,audioinfo.dwLength,data,datasize,&datasize,NULL);
		if (ProcessDataProc(filename,datasize/2)==0)
		{
			free(format);
			free(data);
			CloseHandle(file);
			return E_EABORTED;
		}
	}

	// Now mix them and write the data into the file
	{
		DWORD written;
		long l;

		// RIFF-Kopf
		WriteFile(file,"RIFF",4,&written,NULL);
		l=formatsize+datasize+(4+4+4)+4+4;
		WriteFile(file,&l,sizeof(l),&written,NULL);

		WriteFile(file,"WAVEfmt ",8,&written,NULL);
		l=formatsize;
		WriteFile(file,&l,sizeof(l),&written,NULL);
		
		if (ProcessDataProc(filename,4+sizeof(l)+8+sizeof(l))==0)
		{
			free(format);
			free(data);
			CloseHandle(file);
			return E_EABORTED;
		}

		// Format
		WriteFile(file,format,formatsize,&written,NULL);
		if (written!=DWORD(formatsize))goto error;
		
		if (ProcessDataProc(filename,formatsize/2)==0)
		{
			free(format);
			free(data);
			CloseHandle(file);
			return E_EABORTED;
		}

		// Data head

		WriteFile(file,"data",4,&written,NULL);
		l=datasize;
		WriteFile(file,&l,sizeof(l),&written,NULL);

		// Data
		WriteFile(file,data,datasize,&written,NULL);
		if (DWORD(datasize)!=written)goto error;
		if (ProcessDataProc(filename,datasize/2)==0)
		{
			free(format);
			free(data);
			CloseHandle(file);
			return E_EABORTED;
		}
	}
	free(data);
	free(format);

	CloseHandle(file);

	return 0;
error:
	free(data);
	free(format);
	CloseHandle(file);
	return E_EWRITE;
}

