#include "stdafx.h"
#include "images.h"
#include "ocidl.h"
#include "olectl.h"






#define HIMETRIC_INCH   2540    // HIMETRIC units per inch




HGLOBAL LoadFromFile(PCHAR name,PDWORD lpSize)
{
	// Returns HGLOBAL of JPEG file data
	HGLOBAL memory;
	HANDLE f=CreateFile(name,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,0);
	DWORD e=GetLastError();
	if (f==INVALID_HANDLE_VALUE)return 0;

	*lpSize=GetFileSize(f,NULL);

	memory=GlobalAlloc(GMEM_MOVEABLE,*lpSize);

	LPVOID buf=GlobalLock(memory);

	DWORD read;
	ReadFile(f,buf,*lpSize,&read,NULL);

	GlobalUnlock(memory);

	CloseHandle(f);
	if (read!=*lpSize)
	{
		GlobalFree(memory);
		return 0;
	}
	return memory;
}

HBITMAP LoadJPEG(HINSTANCE lib,PCHAR name,LPWORD width,LPWORD height)
{
	// Extracts JPEG Data and returns a HBITMAP
	IPicture *picture;
	IStream *stream;
	WORD _cx,_cy;
	if (width==NULL)width=&_cx;
	if (height==NULL)height=&_cy;


	HGLOBAL memory=0;
	DWORD size=0;

//	Load Data from file
	if (memory==NULL)memory=LoadFromFile(name,&size);
	if (memory==NULL)return 0;


//	Load Picture from stream
	CoInitialize(NULL);
	CreateStreamOnHGlobal(memory,TRUE,&stream);
	if (stream==NULL)
	{
		GlobalFree(memory);
		CoUninitialize();
		return 0;
	}

	HRESULT hr=OleLoadPicture(stream,0,FALSE,IID_IPicture,(void**)&picture);
	if (hr!=S_OK)
	{
		stream->Release();
		CoUninitialize();
		return 0;
	}

//	Picture loaded, now render to Bitmap
	HBITMAP bitmap;
	{
		OLE_XSIZE_HIMETRIC w,h;

		HDC memdc,desktopdc;

		hr=picture->get_Width(&w);
		hr=picture->get_Height(&h);

		desktopdc=GetDC(0);
		memdc=CreateCompatibleDC(desktopdc);

		{	// Convert units
			DWORD cxPerInch = GetDeviceCaps(memdc,LOGPIXELSX);
			DWORD cyPerInch = GetDeviceCaps(memdc,LOGPIXELSY);
			
			*width=MulDiv(w, cxPerInch, HIMETRIC_INCH);
			*height=MulDiv(h, cyPerInch, HIMETRIC_INCH);
		}

		bitmap=CreateCompatibleBitmap(desktopdc,*width,*height);
		ReleaseDC(0,desktopdc);

		SelectObject(memdc,bitmap);

		hr=picture->Render(memdc,0,0,*width,*height,0,h,w,-h,NULL);

		RestoreDC(memdc,-1);
		DeleteDC(memdc);
	}

//	Clean up	
	picture->Release();
	stream->Release();

	CoUninitialize();
	return bitmap;
}







HBITMAP LoadPicture(PCHAR filename)
{
	// Loads a bitmap file or a JPEG file.
	HBITMAP bitmap=0;
	bitmap=(HBITMAP)LoadImage(0,filename,IMAGE_BITMAP,0,0,LR_LOADFROMFILE);
	if (bitmap!=0)return bitmap;

	bitmap=LoadJPEG(0,filename,NULL,NULL);
	if (bitmap!=0)return bitmap;

	return 0;
}


BOOLEAN SaveBitmap(HBITMAP bitmap,HDC dc,PCHAR filename)
{
	// Saves a bitmap to a file
    int         hFile;
    OFSTRUCT    ofReOpenBuff;
    HBITMAP     hTmpBmp, hBmpOld;
    BOOL        bSuccess;
    BITMAPFILEHEADER    bfh;
    PBITMAPINFO pbmi;
    PBYTE       pBits;
    BITMAPINFO  bmi;
    PBYTE pjTmp, pjTmpBmi;
    ULONG sizBMI;


    bSuccess = TRUE;
    if (!bitmap) return FALSE;

    //
    // Let the graphics engine to retrieve the dimension of the bitmap for us
    // GetDIBits uses the size to determine if it's BITMAPCOREINFO or BITMAPINFO
    // if BitCount != 0, color table will be retrieved
    //
    bmi.bmiHeader.biSize = 0x28;              // GDI need this to work
    bmi.bmiHeader.biBitCount = 0;             // don't get the color table
    if ((GetDIBits(dc, bitmap, 0, 0, (LPSTR)NULL, &bmi, DIB_RGB_COLORS)) == 0) {
       return FALSE;
    }

    //
    // Now that we know the size of the image, alloc enough memory to retrieve
    // the actual bits
    //
    if ((pBits = (PBYTE)GlobalAlloc(GMEM_FIXED | GMEM_ZEROINIT,
                bmi.bmiHeader.biSizeImage)) == NULL) {
        return FALSE;
    }

    //
    // Note: 24 bits per pixel has no color table.  So, we don't have to
    // allocate memory for retrieving that.  Otherwise, we do.
    //
    pbmi = &bmi;                                      // assume no color table

    switch (bmi.bmiHeader.biBitCount) {
        case 24:                                      // has color table
            sizBMI = sizeof(BITMAPINFOHEADER);
            break;
        case 16:
        case 32:
//            sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(DWORD)*3;
			sizBMI = sizeof(BITMAPINFOHEADER);
			bmi.bmiHeader.biCompression=0;
            break;
        default:
            sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*(1<<bmi.bmiHeader.biBitCount);
            break;

    }

    //
    // Allocate memory for color table if it is not 24bpp...
    //
    if (sizBMI != sizeof(BITMAPINFOHEADER)) {
        ULONG       sizTmp;
        //
        // I need more memory for the color table
        //
        if ((pbmi = (PBITMAPINFO)GlobalAlloc(GMEM_FIXED | GMEM_ZEROINIT, sizBMI )) == NULL) {
            bSuccess = FALSE;
            goto ErrExit1;
        }
        //
        // Now that we've a bigger chunk of memory, let's copy the Bitmap
        // info header data over
        //
        pjTmp = (PBYTE)pbmi;
        pjTmpBmi = (PBYTE)&bmi;
        sizTmp = sizeof(BITMAPINFOHEADER);

        while(sizTmp--)
        {
            *(((PBYTE)pjTmp)++) = *((pjTmpBmi)++);
        }
    }

    //
    // Let's open the file and get ready for writing
    //
    if ((hFile = OpenFile(filename, (LPOFSTRUCT)&ofReOpenBuff,
                 OF_CREATE | OF_WRITE)) == -1) {
        goto ErrExit2;
    }

    //
    // But first, fill in the info for the BitmapFileHeader
    //
    bfh.bfType = 0x4D42;                            // 'BM'
    bfh.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+sizBMI+
        pbmi->bmiHeader.biSizeImage;
    bfh.bfReserved1 =
    bfh.bfReserved2 = 0;
    bfh.bfOffBits = sizeof(BITMAPFILEHEADER)+sizBMI;

    //
    // Write out the file header now
    //
    if (_lwrite(hFile, (LPSTR)&bfh, sizeof(BITMAPFILEHEADER)) == -1) {
        bSuccess = FALSE;
        goto ErrExit3;
    }

    //
    // Bitmap can't be selected into a DC when calling GetDIBits
    // Assume that the hDC is the DC where the bitmap would have been selected
    // if indeed it has been selected
    //
    if (hTmpBmp = CreateCompatibleBitmap(dc, pbmi->bmiHeader.biWidth, pbmi->bmiHeader.biHeight)) {
        hBmpOld = (HBITMAP)SelectObject(dc, hTmpBmp);
        if ((GetDIBits(dc, bitmap, 0, pbmi->bmiHeader.biHeight, (LPSTR)pBits, pbmi, DIB_RGB_COLORS))==0){
            bSuccess = FALSE;
            goto ErrExit4;
        }
    } else {
        bSuccess = FALSE;
        goto ErrExit3;
    }

    //
    // Now write out the BitmapInfoHeader and color table, if any
    //
    if (_lwrite(hFile, (LPSTR)pbmi, sizBMI) == -1) {
        bSuccess = FALSE;
        goto ErrExit4;
    }

    //
    // write the bits also
    //
    if (_lwrite(hFile, (LPSTR)pBits, pbmi->bmiHeader.biSizeImage) == -1) {
        bSuccess = FALSE;
        goto ErrExit4;
    }


ErrExit4:
    SelectObject(dc, hBmpOld);
    DeleteObject(hTmpBmp);
ErrExit3:
    _lclose(hFile);
ErrExit2:
    GlobalFree(pbmi);
ErrExit1:
    GlobalFree(pBits);
    return bSuccess;
}

HANDLE  MakeDib( HBITMAP hbitmap, UINT bits )
{
	HANDLE              hdib ;
	HDC                 hdc ;
	BITMAP              bitmap ;
	UINT                wLineLen ;
	DWORD               dwSize ;
	DWORD               wColSize ;
	LPBITMAPINFOHEADER  lpbi ;
	LPBYTE              lpBits ;
	
	GetObject(hbitmap,sizeof(BITMAP),&bitmap) ;

	//
	// DWORD align the width of the DIB
	// Figure out the size of the colour table
	// Calculate the size of the DIB
	//
	wLineLen = (bitmap.bmWidth*bits+31)/32 * 4;
	wColSize = sizeof(RGBQUAD)*((bits <= 8) ? 1<<bits : 0);
	dwSize = sizeof(BITMAPINFOHEADER) + wColSize +
		(DWORD)(UINT)wLineLen*(DWORD)(UINT)bitmap.bmHeight;

	//
	// Allocate room for a DIB and set the LPBI fields
	//
	hdib = GlobalAlloc(GHND,dwSize);
	if (!hdib)
		return hdib ;

	lpbi = (LPBITMAPINFOHEADER)GlobalLock(hdib) ;

	lpbi->biSize = sizeof(BITMAPINFOHEADER) ;
	lpbi->biWidth = bitmap.bmWidth ;
	lpbi->biHeight = bitmap.bmHeight ;
	lpbi->biPlanes = 1 ;
	lpbi->biBitCount = (WORD) bits ;
	lpbi->biCompression = BI_RGB ;
	lpbi->biSizeImage = dwSize - sizeof(BITMAPINFOHEADER) - wColSize ;
	lpbi->biXPelsPerMeter = 0 ;
	lpbi->biYPelsPerMeter = 0 ;
	lpbi->biClrUsed = (bits <= 8) ? 1<<bits : 0;
	lpbi->biClrImportant = 0 ;

	//
	// Get the bits from the bitmap and stuff them after the LPBI
	//
	lpBits = (LPBYTE)(lpbi+1)+wColSize ;

	hdc = CreateCompatibleDC(NULL) ;

	GetDIBits(hdc,hbitmap,0,bitmap.bmHeight,lpBits,(LPBITMAPINFO)lpbi, DIB_RGB_COLORS);

	// Fix this if GetDIBits messed it up....
	lpbi->biClrUsed = (bits <= 8) ? 1<<bits : 0;

	DeleteDC(hdc) ;
	GlobalUnlock(hdib);

	return hdib;
}














































BOOLEAN SaveDIBitmap(LPBITMAPINFOHEADER pdata,PCHAR filename)
{
	// Saves a bitmap to a file
    int         hFile;
    OFSTRUCT    ofReOpenBuff;
    BOOL        bSuccess;
    BITMAPFILEHEADER    bfh;
    ULONG sizBMI;


    bSuccess = TRUE;
    if (!pdata) return FALSE;

    //
    // Let's open the file and get ready for writing
    //
    if ((hFile = OpenFile(filename, (LPOFSTRUCT)&ofReOpenBuff,
                 OF_CREATE | OF_WRITE)) == -1) {
        return FALSE;
    }


    switch (pdata->biBitCount) {
        case 24:                                      // has color table
            sizBMI = sizeof(BITMAPINFOHEADER);
            break;
        case 16:
        case 32:
			if (pdata->biCompression==3)
				sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(DWORD)*3;
			else
				sizBMI = sizeof(BITMAPINFOHEADER);
            break;
        default:
            sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*(1<<pdata->biBitCount);
            break;
    }

    //
    // But first, fill in the info for the BitmapFileHeader
    //
    bfh.bfType = 0x4D42;                            // 'BM'
    bfh.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+sizBMI+
        pdata->biBitCount;
    bfh.bfReserved1 =
    bfh.bfReserved2 = 0;
    bfh.bfOffBits = sizeof(BITMAPFILEHEADER)+sizBMI;

    //
    // Write out the file header now
    //
    if (_lwrite(hFile, (LPSTR)&bfh, sizeof(BITMAPFILEHEADER)) == -1) {
        bSuccess = FALSE;
        goto ErrExit3;
    }

    //
    // Now write out the BitmapInfoHeader, color table and data
    //
    if (_lwrite(hFile, (LPSTR)pdata, sizBMI+pdata->biSizeImage) == -1) {
        bSuccess = FALSE;
    }

ErrExit3:
    _lclose(hFile);
    return bSuccess;
}
