// SerialServer.cpp : Defines the entry point for the application.
//
// Registered "PMTR" code on http://dev.palmos.com/creatorid/
//
// Copyright (C) 2003 Christian Ghisler, C. Ghisler & Co.
// This program is copyrighted freeware

#include "stdafx.h"
#include "resource.h"
#include "sertrans.h"
#include "datatypes.h"
#include "blockio.h"
#include "utils.h"

#define WM_INFORMONCONNECT WM_USER+1

#define numportnames 8
char* portnames[numportnames]={"COM1","COM2","COM3","COM4","COM5","COM6","COM7","COM8"};
#define numspeednames 8
char* speednames[numspeednames]={"115200","57600","38400","28800","19200","9600","4800","2400"};
int speedvalues[numspeednames]={CBR_115200,CBR_57600,CBR_38400,28800,CBR_19200,CBR_9600,CBR_4800,CBR_2400};

// Global Variables:
HINSTANCE hInst;			// current instance
HWND hWndMain;
BOOL inmessageloop=false;
BOOL looprunning=false;
BOOL aborted=false;
BOOL closebuttonpressed=false;
BOOL doreconnect=false;
int	transferbusy=false;	// avoid reentrancy
char lastremoteread[MAX_PATH]={0};
extern int maxblocksize;

char renamefromname[MAX_PATH];

// Foward declarations of functions included in this code module:
ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int,LPSTR);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	About(HWND, UINT, WPARAM, LPARAM);

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
 	// TODO: Place code here.
	MSG msg;
	hInst=hInstance;

	// Initialize global strings
	MyRegisterClass(hInstance);

#ifndef watcom
	InitCommonControls();
#endif

	// Perform application initialization:
	if (!InitInstance (hInstance, nCmdShow,lpCmdLine)) 
	{
		return FALSE;
	}
	// Main message loop:
	while (GetMessage(&msg, NULL, 0, 0)) 
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return msg.wParam;
}



//
//  FUNCTION: MyRegisterClass()
//
//  PURPOSE: Registers the window class.
//
//  COMMENTS:
//
//    This function and its usage is only necessary if you want this code
//    to be compatible with Win32 systems prior to the 'RegisterClassEx'
//    function that was added to Windows 95. It is important to call this function
//    so that the application will get 'well formed' small icons associated
//    with it.
//
ATOM MyRegisterClass(HINSTANCE hInstance)
{
#ifdef watcom
	WNDCLASS wcex;
#define RegisterClassEx RegisterClass
#else
	WNDCLASSEX wcex;
	wcex.cbSize = sizeof(WNDCLASSEX); 
#endif

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= DLGWINDOWEXTRA;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_SERIALSERVER);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_BTNFACE+1);
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= "CGSerialServer";
#ifndef watcom
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);
#endif
	return RegisterClassEx(&wcex);
}

//
//   FUNCTION: InitInstance(HANDLE, int)
//
//   PURPOSE: Saves instance handle and creates main window
//
//   COMMENTS:
//
//        In this function, we save the instance handle in a global variable and
//        create and display the main program window.
//

int tabarray[2]={200,230};

BOOL InitInstance(HINSTANCE hInstance, int nCmdShow,LPSTR lpCmdLine)
{
	HWND hWnd;
	int i;

	hInst = hInstance; // Store instance handle in our global variable

	hWnd = CreateDialog(hInstance,MAKEINTRESOURCE(IDD_MAINDIALOG),0,NULL);

	if (!hWnd)
	{
		return FALSE;
	}

	hWndMain=hWnd;
	char buf[MAX_PATH],ininame[MAX_PATH];
	char* p;
	buf[0]=0;

	GetModuleFileName(hInst,ininame,sizeof(ininame)-1);
	p=strrchr(ininame,'.');
	if (p) {
		p++;
		strcpy(p,"ini");    // Same name as program
	}

	p=NULL;
	GetPrivateProfileString("SerialServer","Settings","",buf,MAX_PATH-1,ininame);

	// command line overrides ini
	if (lpCmdLine && lpCmdLine[0] && (p=strchr(lpCmdLine,' '))) {
#ifndef watcom
		p=lpCmdLine;	// the MSVC++ command line doesn't contain the program name
#endif
		strlcpy(buf,p+1,sizeof(buf)-1);
	}
	p=strchr(buf,' ');
	if (p) {
		p[0]=0;
		p++;
	}

	for (i=0;i<numportnames;i++)
	   SendDlgItemMessage(hWnd,IDC_PORTCOMBO,CB_ADDSTRING,0,(LPARAM)portnames[i]);
	SendDlgItemMessage(hWnd,IDC_PORTCOMBO,CB_SETCURSEL,0,0);
	if (buf[0])
		SendDlgItemMessage(hWnd,IDC_PORTCOMBO,CB_SELECTSTRING,-1,(LPARAM)&buf);

	for (i=0;i<numspeednames;i++)
	   SendDlgItemMessage(hWnd,IDC_SPEEDCOMBO,CB_ADDSTRING,0,(LPARAM)speednames[i]);
	SendDlgItemMessage(hWnd,IDC_SPEEDCOMBO,CB_SETCURSEL,0,0);
	if (p)
		SendDlgItemMessage(hWnd,IDC_SPEEDCOMBO,CB_SELECTSTRING,-1,(LPARAM)p);

	SendDlgItemMessage(hWnd,IDC_REMOTELIST,LB_SETTABSTOPS,2,(LPARAM)&tabarray);

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	return TRUE;
}

BOOL MessageLoop()
{
	MSG msg;
	BOOL retval;
	inmessageloop=true;
	while (PeekMessage(&msg,0,0,0,PM_REMOVE)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	retval=aborted;
	inmessageloop=false;
	aborted=false;
	return retval;
}

BOOL UserAbort()
{
	MSG msg;
	BOOL retval;
	inmessageloop=true;
	while (PeekMessage(&msg,0,0,0,PM_REMOVE)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	retval=aborted;
	inmessageloop=false;
	aborted=false;
	return retval;
}

void ShowStatus(char* status)
{
	SetDlgItemText(hWndMain,IDC_STATUS,status);

}

void ShowName(char* status)
{
	SetDlgItemText(hWndMain,IDC_NAME,status);

}

void UpdatePercentBar(int offset,int filesize)
{
	if (!filesize)
		return;
	int percent=MulDiv(offset,100,filesize);
	if (percent<0) percent=0;
	if (percent>100) percent=100;
#ifdef watcom
	char percentstr[16];
	itoa(percent,percentstr,10);
	strcat(percentstr," %");
	SetDlgItemText(hWndMain,IDC_PROGRESS1,percentstr);
#else
	SendDlgItemMessage(hWndMain,IDC_PROGRESS1,PBM_SETPOS,percent,0);
#endif
}

void SetButtonText(char* txt)
{
	SetDlgItemText(hWndMain,IDC_ABORT,txt);
}

void AddStringToHistory(char* txt)
{
	SendDlgItemMessage(hWndMain,IDC_HISTORY,LB_INSERTSTRING,0,(LPARAM)txt);
}

void AddCommandToHistory(BOOL succeeded,char* cmd,char* txt)
{
	char buf[MAX_PATH];
	char* p=buf;
	if (!succeeded) {
		buf[0]='!';
		p++;
	}
	strlcpy(p,cmd,sizeof(buf)-2);
	strncat(buf," ",sizeof(buf)-1);
	strncat(buf,txt,sizeof(buf)-1);
	AddStringToHistory(buf);
}

int MessageBox(char* txt,int err)
{
	return MessageBox(hWndMain,txt,"Serial Server",err);
}

// ****************** MainReceiveFunction *********************
// *** Main loop to receive and handle commands from client ***
// ************************************************************

void MainReceiveFunction(void)
{
	Err err;
	UInt8 ch;
	HANDLE FileHandle;
	char FileName[260];
	char* p;
	UInt32 attributes;
	FILETIME FileDate;
	UInt32 restartat;
	char* block;
	char cmd;
	BOOL isadir,ok,overwritefile;
	char portname[20];
	int idx, portspeed;

	if (looprunning)		// avoid reentrancy
		return;

	portname[0]=0;
	SetErrorMode(0x08001);  // don't show errors when access to drive fails!

	doreconnect=false;
	restartat=0;
	overwritefile=false;
	block=NULL;

	idx=SendDlgItemMessage(hWndMain,IDC_PORTCOMBO,CB_GETCURSEL,0,0);
	portspeed=SendDlgItemMessage(hWndMain,IDC_SPEEDCOMBO,CB_GETCURSEL,0,0);
	if (idx>=0 && portspeed>=0 && !transferbusy) {
		portspeed=speedvalues[portspeed];
		SendDlgItemMessage(hWndMain,IDC_PORTCOMBO,CB_GETLBTEXT,idx,(LPARAM)&portname);
		SerialConnect(hWndMain,hInst,portname,portspeed,WM_INFORMONCONNECT);
	}

	looprunning=true;

	if (!SerialConnected()) {
		looprunning=false;
	} else if (portname[0]) {      // Save port to ini
		char ininame[MAX_PATH],*p;

		GetModuleFileName(hInst,ininame,sizeof(ininame)-1);
		p=strrchr(ininame,'.');
		if (p) {
			p++;
			strcpy(p,"ini");    // Same name as program
		}

		strcat(portname," ");
		itoa(portspeed,portname+strlen(portname),10);
		WritePrivateProfileString("SerialServer","Settings",portname,ininame);
	}

	SetButtonText("Abort+Disconnect");
	if (looprunning) {
		AddCommandToHistory(true,"Listening on",portname);
		// Try to allocate big buffer - if not possible, use internal(slower)
		block =(char*)malloc(64*1024);
		if (!block) {
			MessageBox("Not enough memory to allocate buffer!",MB_ICONSTOP);
			looprunning=false;
		}

		// Main command loop starts here!

		while (looprunning) {
			ch=RQ_TRANSFERERR;
			ShowStatus("Ready...");
			ShowName("-");

			SerialSendFlush();		// delete any unsent data

			// Wait for next command from server
			FileName[0]=0;
			cmd=ReceiveNextBlock(FileName,sizeof(FileName)-1,NULL,NULL);

			switch(cmd) {
			case CMD_CRCERR:
				err = SerialSendChar(RQ_CRCERR);
				break;
			case CMD_ABORT:				// user-abort
				looprunning=false;		// closes the loop!
				break;
			case CMD_CONNECT:			// sort of ping to check that the server is responding
				ShowStatus("Connect!");
				ch=RQ_OK;
				err=SerialSendChar(ch);
				AddStringToHistory("Connect!");
				break;
			case CMD_LIST:
				isadir=false;
				if (FileName[0]=='\\' && FileName[1]==0) {		// the root
					if (restartat) {
						ShowStatus("Resume not supported!");	// no resume with dirs!
						ch=RQ_TRANSFERERR;
					} else {
						ShowStatus("GET valid...");
						ch=RQ_OK;
					}
					FileHandle=0;
				} else {
					if (strlen(FileName)>4) {  // drives are OK
						HANDLE fh;
						WIN32_FIND_DATA srec;
						if (FileName[strlen(FileName)-1]=='\\')
							FileName[strlen(FileName)-1]=0;
						fh=FindFirstFile(FileName+1,&srec);
						if (fh!=INVALID_HANDLE_VALUE) {
							isadir=srec.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY;
							FindClose(fh);
						}
					} else
						isadir=true;
					if (!isadir) {
						ShowStatus("LIST invalid...");
						ch=RQ_TRANSFERERR;
					} else {
						ShowStatus("LIST valid...");
						ch=RQ_OK;
					}
				}
				err = SerialSendChar(ch);
				if (err==0 && ch==RQ_OK) {
					ShowName(FileName);
					err=SerialSendFile(INVALID_HANDLE_VALUE,FileName,block,restartat,FALSE);
					AddCommandToHistory(err==0,"LIST",FileName);
				} else if (err) {
					strcpy(FileName,"Error ");
					itoa(err,FileName+strlen(FileName),10);
					strcat(FileName," in list");
					AddCommandToHistory(false,"LIST",FileName);
				} else {
					strcat(FileName," (not found)");
					AddCommandToHistory(false,"LIST",FileName);
				}
				break;
			case CMD_GETSIZE:			// client wants to get only file size
			case CMD_GET:				// client tries to get file or dir
				isadir=false;
				if (FileName[0]=='\\' && FileName[1]==0) {		// the root
					ShowStatus("\\ is not a file!...");
					ch=RQ_TRANSFERERR;
					FileHandle=INVALID_HANDLE_VALUE;
				} else {
					FileHandle=CreateFile(FileName+1,GENERIC_READ,FILE_SHARE_READ,
						NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
					if (FileHandle!=INVALID_HANDLE_VALUE && restartat>0) {
						if (SetFilePointer(FileHandle,restartat,NULL,FILE_BEGIN)==-1) {
							CloseHandle(FileHandle);
							FileHandle=INVALID_HANDLE_VALUE;
						}
					}
					if (FileHandle==INVALID_HANDLE_VALUE) {
						ShowStatus("GET invalid...");
						ch=RQ_TRANSFERERR;
					} else {
						ShowStatus("GET valid...");
						ch=RQ_OK;
					}
				}
				err = SerialSendChar(ch);
				if (err==0 && ch==RQ_OK) {
					ShowName(FileName);
					err=SerialSendFile(FileHandle,FileName,block,restartat,cmd==CMD_GETSIZE);
				} else if (err) {
					strcpy(FileName,"Error ");
					itoa(err,FileName+strlen(FileName),10);
					strcat(FileName," in send");
				}
				AddCommandToHistory(err==0 && ch==RQ_OK,cmd==CMD_GETSIZE ? "SIZE" : "GET",FileName);

				if (FileHandle!=INVALID_HANDLE_VALUE)
					CloseHandle(FileHandle);

				FileHandle=INVALID_HANDLE_VALUE;
				break;
			case CMD_SEND:		// client sends file or dir
				if (restartat || overwritefile)
					FileHandle=INVALID_HANDLE_VALUE;
				else
					FileHandle=CreateFile(FileName+1,GENERIC_READ,FILE_SHARE_READ,
						NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
				if (FileHandle!=INVALID_HANDLE_VALUE) {
					CloseHandle(FileHandle);
					FileHandle=INVALID_HANDLE_VALUE;
					ch=RQ_FAILED;
				} else {
					if (restartat)
						FileHandle=CreateFile(FileName+1,GENERIC_WRITE,0,
							NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
					else {
						FileHandle=CreateFile(FileName+1,GENERIC_WRITE,0,
							NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
					}

					if (FileHandle!=INVALID_HANDLE_VALUE) {
						err=0;
						if (restartat) {
							if (SetFilePointer(FileHandle,restartat,NULL,FILE_BEGIN)==-1) {
								AddStringToHistory("Seek for restart failed!");
								err=1;
							} else
								err=0;
						}
					} else
						err=1;

					if (err!=0) {
						ShowStatus("PUT invalid...");
						ch=RQ_TRANSFERERR;
					} else {
						ShowStatus("PUT valid...");
						ch=RQ_OK;
					}
				}

				err = SerialSendChar(ch);
				if (err==0 && ch==RQ_OK) {
					ShowName(FileName);
					err=SerialReceiveFile(FileHandle, FileName, block, restartat);
					AddCommandToHistory(err==0,"PUT",FileName);
				} else if (err) {
					strcpy(FileName,"Error ");
					itoa(err,FileName+strlen(FileName),10);
					strcat(FileName," in recv");
					AddCommandToHistory(false,"PUT",FileName);
				}
				if (FileHandle!=INVALID_HANDLE_VALUE) {
					CloseHandle(FileHandle);
				}
				FileHandle=INVALID_HANDLE_VALUE;
				break;
			case CMD_DELETE:		// delete file or dir
				ShowStatus("Delete file");
				ShowName(FileName);
				ok=DeleteFile(FileName+1);
				if (!ok)
					ok=RemoveDirectory(FileName+1);
				if (ok)
					ch=RQ_OK;
				AddCommandToHistory(ok,"DEL",FileName);
				err = SerialSendChar(ch);
				break;
			case CMD_MKDIR:		// delete file or dir
				if (FileName[0]=='\\' && FileName[2]==':') {
					ShowStatus("Create directory");
					ShowName(FileName);
					ok=CreateDirectory(FileName+1,NULL);
					if (ok)
						ch=RQ_OK;
					AddCommandToHistory(ok,"MKDIR",FileName);
				}
				err = SerialSendChar(ch);
				break;
			case CMD_RENFR:			// renamefrom file or dir
				if (FileName[0]=='\\' && FileName[2]==':' && FileName[3]=='\\') {
					strlcpy(renamefromname,FileName,sizeof(renamefromname)-1);
					ch=RQ_OK;
				}
				AddCommandToHistory(ch==RQ_OK,"RNFR",FileName);
				err = SerialSendChar(ch);
				break;
			case CMD_RENTO:			// rename to file or dir
			case CMD_COPYTO:		// copy to file or dir
				if (renamefromname[0]=='\\' && FileName[0]=='\\' && FileName[2]==':' && FileName[3]=='\\') {
					if (renamefromname[0]) {
						//Server sends RQ_WAIT for long operations
						err = SerialSendChar(RQ_WAIT);
						if (err==0) {
							if (cmd==CMD_COPYTO) {
								if (CopyFile(renamefromname+1,FileName+1,true)) // only same dir!
									ch=RQ_OK;
								else
									ch=RQ_FAILED;
							} else {
								if (MoveFile(renamefromname+1,FileName+1)) // only same dir!
									ch=RQ_OK;
								else
									ch=RQ_FAILED;
							}
						}
					} else
						ch=RQ_FAILED;
				} else
					ch=RQ_FAILED;
				AddCommandToHistory(ch==RQ_OK,cmd==CMD_COPYTO ? "CPTO" : "RNTO",FileName);
				err = SerialSendChar(ch);
				break;
			case CMD_SETATTR:
				p=strchr(FileName,'\t');
				if (p && FileName[0]=='\\' && FileName[2]==':' && FileName[3]=='\\') {
					p[0]=0;
					p++;
					attributes=atoi(p);
					ShowStatus("Change attributes");
					ShowName(FileName);
					if (SetFileAttributes(FileName+1,attributes))
						ch=RQ_OK;
					AddCommandToHistory(ch==RQ_OK,"ATTR",FileName);
				}
				err = SerialSendChar(ch);
				break;
			case CMD_SETTIME:
				p=strchr(FileName,'\t');
				if (p && FileName[0]=='\\' && FileName[2]==':' && FileName[3]=='\\') {
					p[0]=0;
					p++;
					AddCommandToHistory(true,"New time:",p);
					if (ConvertIsoDateToDateTime(p,&FileDate)) {
						ShowStatus("Change date/time");
						ShowName(FileName);
						FileHandle=CreateFile(FileName+1,GENERIC_WRITE,FILE_SHARE_READ,
							NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
						if (FileHandle!=INVALID_HANDLE_VALUE) {
							if (SetFileTime(FileHandle,NULL,NULL,&FileDate))
								ch=RQ_OK;
							AddCommandToHistory(ch==RQ_OK,"TIME",FileName);
							CloseHandle(FileHandle);
						}
					}
				}
				err = SerialSendChar(ch);
				break;
			case CMD_RESTART:
				restartat=atoi(FileName);
				if (restartat>0)
					AddCommandToHistory(true,"RESTART at",FileName);
				ch=RQ_OK;
				err = SerialSendChar(ch);
				break;
			case CMD_OVERWRITE:
				overwritefile=1;
				AddStringToHistory("Overwrite enabled");
				ch=RQ_OK;
				err = SerialSendChar(ch);
				break;
			default:
				ch=RQ_FAILED;
				err = SerialSendChar(ch);
			}
			if (cmd!=CMD_RESTART && cmd!=CMD_CRCERR)	// only use restart for next command!
				restartat=0;
			if (cmd!=CMD_OVERWRITE && cmd!=CMD_CRCERR)	// only use overwrite for next command!
				overwritefile=0;
		}
	}
	// Change the status text.

	ShowStatus("Finished");
	SetButtonText("Connect");

	if (block)
		free(block);
	SerialCloseConnection();
	if (closebuttonpressed)
		PostMessage(hWndMain,WM_SYSCOMMAND,SC_CLOSE,0);

	looprunning=false;
}

//
//  FUNCTION: WndProc(HWND, unsigned, WORD, LONG)
//
//  PURPOSE:  Processes messages for the main window.
//
//  WM_COMMAND	- process the application menu
//  WM_PAINT	- Paint the main window
//  WM_DESTROY	- post a quit message and return
//
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	
	switch (message) 
	{
		case WM_SHOWWINDOW:
			PostMessage(hWnd,WM_INFORMONCONNECT,0,0);
			return DefWindowProc(hWnd, message, wParam, lParam);
		case WM_SYSCOMMAND:
			if (inmessageloop && wParam==SC_CLOSE) {
				aborted=true;
				closebuttonpressed=true;
				return 0;
			}
			return DefWindowProc(hWnd, message, wParam, lParam);
		case WM_COMMAND:
			wmId    = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			// Parse the menu selections:
			switch (wmId)
			{
				case IDM_EXIT:
					DestroyWindow(hWnd);
					break;
				case IDC_ABORT:
					if (inmessageloop) {
						aborted=true;
					} else
						PostMessage(hWnd,WM_INFORMONCONNECT,0,0);
					break;
				case IDC_PORTCOMBO:
				case IDC_SPEEDCOMBO:
					if (wmEvent==CBN_CLOSEUP && !transferbusy) {
						doreconnect=true;
						aborted=true;
					}
					break;
				default:
					return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;
		case WM_INFORMONCONNECT:
			do {
				aborted=false;
				MainReceiveFunction();
			} while (doreconnect);
			break;
		case WM_DESTROY:
			SerialCloseConnection();
			PostQuitMessage(0);
			return 0;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

