/*****************************************************************************
**
**  FILE   :	UIFunc.cpp
**
**  PURPOSE:	Module contains user interface support functions
**
**	Author:		Konstantin Polyakov
**
*****************************************************************************/

#include "GlobalVars.h"
#include "UIFunc.h"
#include "commctrl.h"
#include "resource.h" 
#include "stdio.h" 



/*****************************************************************************
	Routine:     CheckButton
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

void CheckButton( HWND hDlg, UINT ItemCode )
{
	HWND hItem =  GetDlgItem( 
				hDlg, 
				ItemCode
				);

	SendMessage(  
		hItem, 
		BM_SETCHECK,
		BST_CHECKED, 
		0
		);
}

/*****************************************************************************
	Routine:     UncheckButton
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

void UncheckButton( HWND hDlg, UINT ItemCode )
{
	HWND hItem =  GetDlgItem( 
				hDlg, 
				ItemCode
				);

	SendMessage(  
		hItem, 
		BM_SETCHECK,
		BST_UNCHECKED, 
		0
		);
}

/*****************************************************************************
	Routine:     EnableControl
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

void EnableControl( HWND hDlg, UINT ItemCode )
{
	HWND hItem =  GetDlgItem( 
				hDlg, 
				ItemCode
				);

	EnableWindow( hItem, TRUE );
}

/*****************************************************************************
	Routine:     EnableControl
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

void DisableControl( HWND hDlg, UINT ItemCode )
{
	HWND hItem =  GetDlgItem( 
				hDlg, 
				ItemCode
				);

	EnableWindow( hItem, FALSE );
}

/*****************************************************************************
	Routine:     ChildDialogProc
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

BOOL CALLBACK ChildDialogProc(
			HWND hChildDlg, 
			UINT message, 
			WPARAM wParam, 
			LPARAM lParam
			)
{
	switch ( message )
	{
		case WM_INITDIALOG:

			RECT rcTab, rc;

			GetWindowRect( g_DlgDesc.hTabCtrl, &rcTab );
			TabCtrl_AdjustRect( g_DlgDesc.hTabCtrl, FALSE, &rcTab );
			CopyRect( &rc, &rcTab );
			OffsetRect( &rc, -rc.left, -rc.top );

			SetWindowPos(
				hChildDlg, 
				HWND_TOP, 
				rcTab.left, 
				rcTab.top, 
				rc.right, 
				rc.bottom, 
				0
				); 
			
			return TRUE;


		case WM_COMMAND:

			LRESULT	rv;
			HWND	hItem;

			switch  ( wParam )
			{
				case IDC_CHK_DIRECTORIES:

					hItem = GetDlgItem( 
								hChildDlg, 
								IDC_CHK_DIRECTORIES
								);

					rv = SendMessage(  
							hItem, 
							BM_GETCHECK,
							0, 
							0
							);

					if ( rv == BST_UNCHECKED )
						CheckButton(
							g_DlgDesc.hChildDlg[0], 
							IDC_CHK_FILE_NAME
							);

				case IDC_CHK_FILE_NAME:

					hItem = GetDlgItem( 
								hChildDlg, 
								IDC_CHK_FILE_NAME
								);

					rv = SendMessage(  
							hItem, 
							BM_GETCHECK,
							0, 
							0
							);

					if ( rv == BST_UNCHECKED )
					{
						DisableControl(
								hChildDlg, 
								IDC_CHK_FULL_NAME
								);

						DisableControl(
								hChildDlg, 
								IDC_CHK_SIZE
								);

						DisableControl(
								hChildDlg, 
								IDC_CHK_DATE
								);

						DisableControl(
								hChildDlg, 
								IDC_CHK_TIME
								);

						DisableControl(
								hChildDlg, 
								IDC_CHK_EXT
								);

						DisableControl(
								hChildDlg, 
								IDC_CHK_ATTR
								);

						DisableControl(
								hChildDlg, 
								IDC_ED_FILE_TYPES
								);

						CheckButton(
							g_DlgDesc.hChildDlg[0], 
							IDC_CHK_DIRECTORIES
							);
					}
					else 
					if ( rv == BST_CHECKED )
					{
						EnableControl(
								hChildDlg, 
								IDC_CHK_FULL_NAME
								);

						EnableControl(
								hChildDlg, 
								IDC_CHK_SIZE
								);

						EnableControl(
								hChildDlg, 
								IDC_CHK_DATE
								);

						EnableControl(
								hChildDlg, 
								IDC_CHK_TIME
								);

						EnableControl(
								hChildDlg, 
								IDC_CHK_EXT
								);

						EnableControl(
								hChildDlg, 
								IDC_CHK_ATTR
								);

						EnableControl(
								hChildDlg, 
								IDC_ED_FILE_TYPES
								);
					}

					return TRUE;

			}

			break;
	}

    return FALSE;
}

/*****************************************************************************
	Routine:     OnSelChanged
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

VOID OnSelChanged( HWND hDlg ) 
{ 
	UINT CurPageIndex = TabCtrl_GetCurSel( g_DlgDesc.hTabCtrl ); 

	if ( g_DlgDesc.hCurDlg != NULL ) 
        ShowWindow( g_DlgDesc.hCurDlg, SW_HIDE ); 
 
    // Create the new child dialog box. 
    g_DlgDesc.hCurDlg = g_DlgDesc.hChildDlg[CurPageIndex];
	ShowWindow( 
		g_DlgDesc.hCurDlg, 
		SW_SHOW 
		);
} 

/*****************************************************************************
	Routine:     DoLockDlgRes
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

DLGTEMPLATE * DoLockDlgRes( LPCTSTR lpszResName ) 
{ 
    HRSRC hrsrc		= FindResource( g_hModuleDll, lpszResName, RT_DIALOG ); 
    HGLOBAL hglb	= LoadResource( g_hinst, hrsrc ); 

    return (DLGTEMPLATE *) LockResource( hglb ); 
}

/*****************************************************************************
	Routine:     OnInitOptionDialog
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

void OnInitOptionDialog( HWND hDlg )
{
	TCITEM	Attributes; 

	// Create Tab Control
	g_DlgDesc.hTabCtrl = GetDlgItem( hDlg, IDC_TAB1 );

	Attributes.mask = TCIF_TEXT | TCIF_IMAGE; 
	Attributes.iImage = -1; 
	Attributes.pszText = "View";
	TabCtrl_InsertItem( g_DlgDesc.hTabCtrl, 0, &Attributes ); 
	Attributes.pszText = "Sort"; 
	TabCtrl_InsertItem( g_DlgDesc.hTabCtrl, 1, &Attributes ); 
	Attributes.pszText = "Format"; 
	TabCtrl_InsertItem( g_DlgDesc.hTabCtrl, 2, &Attributes ); 
	Attributes.pszText = "About"; 
	TabCtrl_InsertItem( g_DlgDesc.hTabCtrl, 3, &Attributes ); 

	// load and create dialogs
	for ( int i = 0; i < OPTION_PAGES; i++ )
	{
        g_DlgDesc.ChildDlgRes[i] = 
			DoLockDlgRes( MAKEINTRESOURCE( IDD_VIEWDLG+i ) ); 

		// Create the new child dialog box. 
		g_DlgDesc.hCurDlg = 
		g_DlgDesc.hChildDlg[i] =
			CreateDialogIndirect(
					g_hinst,
					g_DlgDesc.ChildDlgRes[i],
					hDlg,
					ChildDialogProc
					);
	}

	// setup default parameters

	// --------- for VIEW dialog --------------
	
	if ( g_ViewParam.bDirName )
		CheckButton(
			g_DlgDesc.hChildDlg[0], 
			IDC_CHK_DIRECTORIES
			);

	if ( g_ViewParam.bFileName )
		CheckButton(
			g_DlgDesc.hChildDlg[0], 
			IDC_CHK_FILE_NAME 
			);
	else
	{
		HWND hChildDlg = g_DlgDesc.hChildDlg[0];

		DisableControl(
				hChildDlg, 
				IDC_CHK_FULL_NAME
				);

		DisableControl(
				hChildDlg, 
				IDC_CHK_SIZE
				);

		DisableControl(
				hChildDlg, 
				IDC_CHK_DATE
				);

		DisableControl(
				hChildDlg, 
				IDC_CHK_TIME
				);

		DisableControl(
				hChildDlg, 
				IDC_CHK_EXT
				);

		DisableControl(
				hChildDlg, 
				IDC_CHK_ATTR
				);

		DisableControl(
				hChildDlg, 
				IDC_ED_FILE_TYPES
				);
	}

	if ( g_ViewParam.bFullName )
		CheckButton(
			g_DlgDesc.hChildDlg[0], 
			IDC_CHK_FULL_NAME
			);

	if ( g_ViewParam.bExt )
		CheckButton(
			g_DlgDesc.hChildDlg[0], 
			IDC_CHK_EXT
			);

	if ( g_ViewParam.bSize )
		CheckButton(
			g_DlgDesc.hChildDlg[0], 
			IDC_CHK_SIZE
			);

	if ( g_ViewParam.bDate )
		CheckButton(
			g_DlgDesc.hChildDlg[0], 
			IDC_CHK_DATE
			);

	if ( g_ViewParam.bTime )
		CheckButton(
			g_DlgDesc.hChildDlg[0], 
			IDC_CHK_TIME
			);

	if ( g_ViewParam.bAttr )
		CheckButton(
			g_DlgDesc.hChildDlg[0], 
			IDC_CHK_ATTR
			);

	HWND hItem =  GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_ED_FILE_TYPES
				);

	SendMessage(  
		hItem, 
		WM_SETTEXT,
		0, 
		(LPARAM)g_ViewParam.sFileTypes
		);

	// --------- for SORT dialog --------------

	if ( g_SortParam.bName )
		CheckButton(
			g_DlgDesc.hChildDlg[1], 
			IDC_RB_SORT_NAME
			);

	if ( g_SortParam.bExt )
		CheckButton(
			g_DlgDesc.hChildDlg[1], 
			IDC_RB_SORT_EXT
			);

	if ( g_SortParam.bSize )
		CheckButton(
			g_DlgDesc.hChildDlg[1], 
			IDC_RB_SORT_SIZE
			);

	if ( g_SortParam.bDate )
		CheckButton(
			g_DlgDesc.hChildDlg[1], 
			IDC_RB_SORT_DATE
			);

	if ( g_SortParam.bDescent )
		CheckButton(
			g_DlgDesc.hChildDlg[1], 
			IDC_CHK_SORT_DESC
			);


	// --------- for FORMAT dialog --------------

	if ( g_FormatParam.bIndentFiles )
		CheckButton(
			g_DlgDesc.hChildDlg[2], 
			IDC_CHK_FORMAT_FILES
			);

	if ( g_FormatParam.bIndentAll )
		CheckButton(
			g_DlgDesc.hChildDlg[2], 
			IDC_CHK_FORMAT_ALL
			);

	if ( g_FormatParam.bExtSeparately )
		CheckButton(
			g_DlgDesc.hChildDlg[2], 
			IDC_CHK_FORMAT_EXT
			);

	hItem =  GetDlgItem( 
				g_DlgDesc.hChildDlg[2], 
				IDC_ED_FORMAT_WIDTH
				);

	char Width[10];

	_itoa( g_FormatParam.Width, Width, 10 );

	SendMessage(  
		hItem, 
		WM_SETTEXT,
		(WPARAM) 0, 
		(LPARAM) Width
		);
}

/*****************************************************************************
	Routine:     GetOptionsFromWindow
------------------------------------------------------------------------------
	Description: Read options from the options dialog into global variables
  
	Arguments:  

	Return Value:


*****************************************************************************/

void GetOptionsFromWindow()
{
	UINT	length;

	// -------- View Dialog -----------------

	HWND hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_CHK_DIRECTORIES
				);

	LRESULT rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_ViewParam.bDirName = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_CHK_FILE_NAME
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_ViewParam.bFileName = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_CHK_FULL_NAME
				);

	rv = SendMessage( hItem, BM_GETCHECK, 0, 0 );

	g_ViewParam.bFullName = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_CHK_EXT
				);

	rv = SendMessage( hItem, BM_GETCHECK, 0, 0 );

	g_ViewParam.bExt = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_CHK_SIZE
				);

	rv = SendMessage( hItem, BM_GETCHECK, 0, 0 );

	g_ViewParam.bSize = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_CHK_DATE
				);

	rv = SendMessage( hItem, BM_GETCHECK, 0, 0 );

	g_ViewParam.bDate = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_CHK_TIME
				);

	rv = SendMessage( hItem, BM_GETCHECK, 0, 0 );

	g_ViewParam.bTime = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_CHK_ATTR
				);

	rv = SendMessage( hItem, BM_GETCHECK, 0, 0 );

	g_ViewParam.bAttr = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem =  GetDlgItem( 
				g_DlgDesc.hChildDlg[0], 
				IDC_ED_FILE_TYPES
				);

	SendMessage(  
		hItem, 
		WM_GETTEXT,
		(WPARAM)&length, 
		(LPARAM)g_ViewParam.sFileTypes
		);

	// -------- Sort Dialog -----------------

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[1], 
				IDC_RB_SORT_NAME
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_SortParam.bName = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[1], 
				IDC_RB_SORT_EXT
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_SortParam.bExt = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[1], 
				IDC_RB_SORT_SIZE
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_SortParam.bSize = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[1], 
				IDC_RB_SORT_DATE
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_SortParam.bDate = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[1], 
				IDC_CHK_SORT_DESC
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_SortParam.bDescent = 
		rv == BST_CHECKED ? TRUE : FALSE;

	// -------- Format Dialog -----------------

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[2], 
				IDC_CHK_FORMAT_FILES
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_FormatParam.bIndentFiles = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[2], 
				IDC_CHK_FORMAT_ALL
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_FormatParam.bIndentAll = 
		rv == BST_CHECKED ? TRUE : FALSE;

	///////////////////////////////////////

	hItem =  GetDlgItem( 
				g_DlgDesc.hChildDlg[2], 
				IDC_ED_FORMAT_WIDTH
				);

	char sWidth[10];

	SendMessage(  
		hItem, 
		WM_GETTEXT,
		(WPARAM)&length, 
		(LPARAM)sWidth
		);

	g_FormatParam.Width = atoi( sWidth );

	///////////////////////////////////////

	hItem = GetDlgItem( 
				g_DlgDesc.hChildDlg[2], 
				IDC_CHK_FORMAT_EXT
				);

	rv = 
		SendMessage( hItem, BM_GETCHECK, 0,	0 );

	g_FormatParam.bExtSeparately = 
		rv == BST_CHECKED ? TRUE : FALSE;

}

/*****************************************************************************
	Routine:     OptionsDialogProc
------------------------------------------------------------------------------
	Description: 
  
	Arguments:  

	Return Value:


*****************************************************************************/

BOOL CALLBACK OptionsDialogProc(
			HWND hDlg, 
			UINT message, 
			WPARAM wParam, 
			LPARAM lParam
			)
{
	switch ( message )
	{
		case WM_INITDIALOG:
			{
			
			SetCurrentDirectory( g_pWorkingDir );

			//FILE *OptionsFile = fopen( "CatalogMaker.ini", "r" );
			FILE *OptionsFile = fopen( g_pCfgFileName, "r" );

			if ( OptionsFile )
			{
				fread( 
					&g_ViewParam, 
					sizeof(g_ViewParam), 
					1, 
					OptionsFile 
				);

				fread( 
					&g_SortParam, 
					sizeof(g_SortParam), 
					1, 
					OptionsFile 
				);

				fread( 
					&g_FormatParam, 
					sizeof(g_FormatParam), 
					1, 
					OptionsFile 
				);

				fclose( OptionsFile );
			}

			g_MainWin = hDlg;
			OnInitOptionDialog( hDlg );
			OnSelChanged( hDlg );
			return TRUE;
			}

		case WM_MOVE:

			for ( int i = 0; i < OPTION_PAGES; i++ )
			{
				SendMessage( g_DlgDesc.hChildDlg[i], WM_INITDIALOG, 0, 0 );
			}

			break;

		case WM_NOTIFY:
			
			NMHDR *p;
			p = (NMHDR*) lParam;

			switch ( p->code ) 
			{ 
				case TCN_SELCHANGE:

						OnSelChanged( hDlg );
						return TRUE;
            } 
            break; 

		case WM_COMMAND:
		
            switch ( wParam )
			{
				case IDOK:
				{
					GetOptionsFromWindow();
					SetCurrentDirectory( g_pWorkingDir );

					//FILE *OptionsFile = fopen( "CatalogMaker.ini", "w" );
					FILE *OptionsFile = fopen( g_pCfgFileName, "w" );

					if ( OptionsFile )
					{
						fwrite( 
							&g_ViewParam, 
							sizeof(g_ViewParam), 
							1, 
							OptionsFile 
						);

						fwrite( 
							&g_SortParam, 
							sizeof(g_SortParam), 
							1, 
							OptionsFile 
						);

						fwrite( 
							&g_FormatParam, 
							sizeof(g_FormatParam), 
							1, 
							OptionsFile 
						);

						fclose( OptionsFile );
					}
					else
					{
						MessageBox(
							NULL,
							"Inpossible to write options on your hard disk",
							"Warning",
							MB_OK
						);
					}
				}

				case IDCANCEL:

					EndDialog( hDlg, wParam );
					return TRUE;

			}
			break;
	}

    return FALSE;
}
