// ************************************************************
// ********************** FilePCtoPDA.c ***********************
// ************************************************************
// **** Copyright 2003 Christian Ghisler, C. Ghisler & Co. ****
// ****************** All rights reserved *********************
// ************ Handspring code added by aziemer **************
// ************************************************************

// ******************* System header files ********************
// ************************************************************

#include <PalmOS.h>
#include <DataMgr.h>			// Access to database functions
#include <VFSMgr.h>				// Access to SD card functions
#include <SerialMgrOld.h>		// Old serial manager
#include <SerialMgr.h>			// New serial manager

#include "filepc2pda.h"
#include "filepc2pda_res.h"
#include "mbox.h"

// ******************** Global constants **********************
// ************************************************************

#define ThisPrograms4byteCode 'PMTR'
#define historylen 100

#define mainheadersize 78
#define pdbrecheadersize 8
#define prcrecheadersize 10

#define memmove MemMove

#define portspeedlistcount 9
Int32 portspeedlist[]={-1,1,115200,57600,38400,28800,19200,9600,4800,2400};


// ******************** Global variables **********************
// ************************************************************

// Version info
UInt32 romversion;

// Current disk volume
UInt16 VolRefNumCurrent=0;
UInt16 VolRefNumCount=0;
UInt16 VolRefNumList[10];  // max 10 volumes supported

// Old name for renaming
char renamefromname[260]={0};

char disconnectstr[]="Please disconnect first!";

// UI stuff
FormType *mainform;
BOOL looprunning=false;
char* historylist[historylen]={0};
UInt32 historycount=0;

// Serial stuff
UInt32 maxblockpart;
UInt32 maxblocksize=defmaxblocksize;
UInt16 SerialPortHandle=0;
UInt32 SerialSpeed=115200;		// is remembered across sessions!
BOOL isUSB=true;
BOOL isBlueTooth=false;
BOOL NewSerialManagerPresent=true;
BOOL UseVersion1SendRecv=false;
BOOL wantromdatabases=false;
BOOL waitingforcommand=false;
BOOL continuesleepneeded=false;

#define usbLibSrmDriverIDDesktop	'HsUd'
BOOL isHandspring=false;
BOOL isSony=false;
BOOL hasBlueToothSupport=false;

// *********************** StrLCopy ***************************
// ******* Pascal function to copy max maxlen chars, **********
// **************** but WITH trailing space! ******************
// ************************************************************

char* StrLCopy(char* p,const char* p2,int maxlen)
{
	if ((int)StrLen(p2)>=maxlen) {
		StrNCopy(p,p2,maxlen);
		p[maxlen]=0;
	} else
		StrCopy(p,p2);
	return p;
}

// ************************ StrRChr ***************************
// *** Function to search for last occurance of a character ***
// ************************************************************

char* StrRChr(char* p,char ch)
{
	char* pout;
	pout=p+StrLen(p);
	while (pout>p) {
		pout--;
		if (pout[0]==ch)
			return pout;
	}
	return NULL;
}

// ******************* ISO-Date functions *********************
// ****** Convert from/to ISO date format: YYYYMMDDhhmmss *****
// ************************************************************

int conv2bytes(char* p)
{
	char buf[16];
	StrLCopy(buf,p,2);
	return StrAToI(buf);
}

UInt32 ConvertIsoDateToDateTime(char* pdatetimefield)
{
	DateTimeType datetime;
	char buf[16];
	StrLCopy(buf,pdatetimefield,4);
	datetime.year=StrAToI(buf);
	datetime.month=conv2bytes(pdatetimefield+4);
	datetime.day=conv2bytes(pdatetimefield+6);
	datetime.hour=conv2bytes(pdatetimefield+8);
	datetime.minute=conv2bytes(pdatetimefield+10);
	datetime.second=conv2bytes(pdatetimefield+12);
	datetime.weekDay=0;
	return TimDateTimeToSeconds(&datetime);
}

void Conv2Chars(char* buf,UInt16 nr)
{
	if (nr<=9) {
		buf[0]='0';
		buf[1]='0'+nr;
	} else {
		StrIToA(buf,nr);
	}
}

void CreateIsoDateString(UInt32 modDate,char* buf)
{
	DateTimeType datetime;
	TimSecondsToDateTime(modDate,&datetime);
	StrIToA(buf,datetime.year);
	Conv2Chars(buf+4,datetime.month);
	Conv2Chars(buf+6,datetime.day);
	Conv2Chars(buf+8,datetime.hour);
	Conv2Chars(buf+10,datetime.minute);
	Conv2Chars(buf+12,datetime.second);
	buf[14]=0;
}

// ************************ Sleep *****************************
// ********** Sleep the given number of milliseconds **********
// ******* Equivalent of Windows function with same name ******
// ************************************************************

void Sleep(UInt32 ms)
{
	SysTaskDelay(PalmSysTicksPerSecond()*ms/1000);
}

// ********************** PilotMain ***************************
// *************** Main program entry point *******************
// **** Sample function taken from "Palm OS Companion.pdf" ****
// ************************************************************

UInt32 PilotMain (UInt16 cmd,void *cmdPBP,UInt16 launchFlags)
{
	UInt16 error;
	UInt16 cardNo;
	LocalID dbID;
	UInt32 notificationmgrpresent;
	SysNotifyParamType* sep;

	if (cmd == sysAppLaunchCmdNormalLaunch){
		error=StartApplication();
		if (error)return(error);

		if (FtrGet(sysFtrCreator,sysFtrNumNotifyMgrVersion,&notificationmgrpresent))
			notificationmgrpresent=0;

		// Register for sleep notification - prevents crash when uses presses power button during a transfer
		if (notificationmgrpresent) {
			SysCurAppDatabase(&cardNo,&dbID);
			SysNotifyRegister(cardNo,dbID,sysNotifySleepRequestEvent,NULL,sysNotifyNormalPriority,NULL);
		}

		FrmGotoForm(ID_MAINFORM);
		EventLoop();

		if (notificationmgrpresent)
			SysNotifyUnregister(cardNo,dbID,sysNotifySleepRequestEvent,sysNotifyNormalPriority);
		StopApplication();
	} else if (cmd == sysAppLaunchCmdNotify) {
		sep=(SysNotifyParamType*)cmdPBP;
		if (sep->notifyType==sysNotifySleepRequestEvent) {
			if (looprunning) {
				if (waitingforcommand) {   // idle waiting for next command -> abort+close
					EventType event;
					event.eType=ctlSelectEvent;
					event.data.ctlEnter.controlID=ID_STARTSTOP;
					event.data.ctlEnter.pControl=NULL;
					EvtAddEventToQueue(&event);
				} else {
					AddStringToHistory("Power off delayed (file transfer)");
					looprunning=false;   // end the loop after current command ends
				}
				continuesleepneeded=true;
				((SleepEventParamType*)sep->notifyDetailsP)->deferSleep=1;
			}
		}		
	}
	return(0);
}

// ****************** StartApplication ************************
// **************** Load user preferences *********************
// **** Sample function taken from "Palm OS Companion.pdf" ****
// ************************************************************

typedef struct _Filepc2pdaPreferenceType {
	Int32 SerialSpeed;
	BOOL wantromdatabases;
} Filepc2pdaPreferenceType;

#define PortPrefID 1
#define PortPrefVersion 1

#define btLibFeatureCreator      sysFileCBtLib
#define btLibFeatureVersion      0 // feature number to get version from the 'blth' feature 
#define sonyHwrOEMCompanyID_Sony      'sony'

static UInt16 StartApplication(void)
{
	Filepc2pdaPreferenceType prefs;
	UInt32 value,btVersion;
	UInt16 prefsSize;
	BOOL prefloaded;

	if (FtrGet(sysFtrCreator,sysFtrNumROMVersion,&romversion))
		return 1;

	// From Application note 09 - "Data Communications on Handspring Devices"
	if (!FtrGet('hsEx',0,&value))
		isHandspring=true;

	if(!FtrGet(sysFtrCreator, sysFtrNumOEMCompanyID, &value))
		if (value == sonyHwrOEMCompanyID_Sony)
			isSony=true;

	// From Palm OS Companion 2.pdf
	hasBlueToothSupport=FtrGet(btLibFeatureCreator, btLibFeatureVersion,&btVersion) == errNone;
	
	// Get last port settings here!
	prefsSize=sizeof(Filepc2pdaPreferenceType);
	if (sysGetROMVerMajor(romversion) >= 2)
		prefloaded=(PrefGetAppPreferences(ThisPrograms4byteCode,
			PortPrefID,&prefs,&prefsSize,true)!=noPreferenceFound);
	else
		prefloaded=PrefGetAppPreferencesV10(ThisPrograms4byteCode,
			PortPrefVersion,&prefs,prefsSize);

	if (prefloaded){
		wantromdatabases = prefs.wantromdatabases;
		SerialSpeed = prefs.SerialSpeed;
		if (SerialSpeed==1) {
			SerialSpeed=115200;
			isBlueTooth=hasBlueToothSupport;
			isUSB=false;
		} else if (SerialSpeed==-1) {
			SerialSpeed=115200;
			isBlueTooth=false;
			isUSB=true;
		} else {
			isBlueTooth=false;
			isUSB=false;
		}
	} else if (sysGetROMVerMajor(romversion) <= 3) {	// default to serial!
		isUSB=false;
		isBlueTooth=false;
	}		
	return 0;
}

// ******************* StopApplication ************************
// ********** Save user preferences, close forms **************
// **** Sample function taken from "Palm OS Companion.pdf" ****
// ************************************************************

static void StopApplication(void)
{
	ControlType* romcheckbox;
	Filepc2pdaPreferenceType prefs;
	UInt16 prefsSize;
	int i;

	// Save last port settings here!
	if (isBlueTooth)
		prefs.SerialSpeed=1;
	else if (isUSB)
		prefs.SerialSpeed=-1;
	else
		prefs.SerialSpeed=SerialSpeed;

	romcheckbox=FrmGetObjectPtr(mainform,FrmGetObjectIndex(mainform,ID_ROM));
	if (romcheckbox)
		wantromdatabases=CtlGetValue(romcheckbox)!=0;
		prefs.wantromdatabases=wantromdatabases;

	prefsSize=sizeof(Filepc2pdaPreferenceType);
	if (sysGetROMVerMajor(romversion) >= 2)
		PrefSetAppPreferences(ThisPrograms4byteCode,PortPrefID,
			PortPrefVersion,&prefs,prefsSize,true);
	else
		PrefSetAppPreferencesV10(ThisPrograms4byteCode,
			PortPrefVersion,&prefs,prefsSize);

	for (i=0;i<historycount;i++)
		if (historylist[i])
			MemPtrFree(historylist[i]);

	//Close all the open forms.
	FrmCloseAllForms();
}

// ********************** EventLoop ***************************
// ********* Handle events while program is idle **************
// **** Sample function taken from "Palm OS Companion.pdf" ****
// ************************************************************

static void EventLoop(void)
{
	UInt16 error;
	EventType event;
	do
	{
		EvtGetEvent(&event,evtWaitForever);
		if (!SysHandleEvent(&event))
			if (!MenuHandleEvent(NULL,&event,&error))
				if (!ApplicationHandleEvent(&event))
					FrmDispatchEvent(&event);
		#if EMULATION_LEVEL !=EMULATION_NONE
		ECApptDBValidate(ApptDB);
		#endif
	}
	while (event.eType !=appStopEvent);
}

// **************** ApplicationHandleEvent ********************
// ********* Handle own events while program is idle **********
// ************************************************************

BOOL ApplicationHandleEvent(EventType *event)
{
	FormType *form;
	ListType *lst;

	switch(event->eType) {
	case frmLoadEvent:
		form=FrmInitForm(event->data.frmLoad.formID);
		mainform=form;
		FrmSetActiveForm(form);
		FrmSetEventHandler(form, MainFormEventHandler);
		lst=FrmGetObjectPtr(mainform,FrmGetObjectIndex(mainform, ID_HISTORY));
		if (lst) {
			LstSetDrawFunction(lst,ListDrawFunction);
		}
		break;
	default:
		break;
	}
	return false;
}

// ***************** SetPortButtonAndHistory ******************
// ******** Set the label of the change port button ***********
// ************************************************************

void SetPortButtonAndHistory()
{
	char buf[30];

	StrCopy(buf,"ROM version: 0x");
	StrIToH(buf+StrLen(buf),romversion);
	AddStringToHistory(buf);

	StrCopy(buf,"Conf. Port: ");
	if (isBlueTooth)
		StrCat(buf,"Bluetooth");
	else if (isUSB)
		StrCat(buf,"USB");
	else
		StrIToA(buf+StrLen(buf),SerialSpeed);
	AddStringToHistory(buf);
	SetPortButtonText(buf+6);
}

// ****************** MainFormEventHandler ********************
// ********* Handle events while program is idle **************
// ************************************************************

Boolean MainFormEventHandler(EventType *event)
{
	ControlType* romcheckbox;
	switch(event->eType) {
	case frmOpenEvent:
		FrmDrawForm(FrmGetActiveForm());
		SetPortButtonAndHistory();
		romcheckbox=FrmGetObjectPtr(mainform,FrmGetObjectIndex(mainform,ID_ROM));
		if (romcheckbox)
			CtlSetValue(romcheckbox,wantromdatabases ? 1 : 0);
		MainReceiveFunction();
		break;

	case ctlSelectEvent:
		switch(event->data.ctlEnter.controlID) {
		case ID_STARTSTOP:
			MainReceiveFunction();
			return true;
		case ID_CONFIGURE:
			if (ShowConfigurePortForm()) {
				SetPortButtonAndHistory();
			}
			return true;
		}
	default:
		break;
	}
	return false;
}

// ****************** ShowConfigurePortForm *******************
// ********* Displays form to configure the com port: *********
// ************ Does not need its own window proc! ************
// ************************************************************

BOOL ShowConfigurePortForm()
{
	FormType *portform;
	ListType *lst;
	int i;
	BOOL retval;
	retval=false;
	portform=FrmInitForm(ID_PORTFORM);
	if (!portform)
		return false;

	lst=FrmGetObjectPtr(portform,FrmGetObjectIndex(portform, ID_PORTSPEED));
	if (lst) {
		if (isUSB)
			LstSetSelection(lst,0);
		else if (isBlueTooth)
			LstSetSelection(lst,1);
		else for (i=0;i<portspeedlistcount;i++)
			if (SerialSpeed==portspeedlist[i]) {
				LstSetSelection(lst,i);
				break;
			}
	}
	FrmSetActiveForm(portform);
	FrmDrawForm(portform);
	if (1==FrmDoDialog(portform) && lst) {
		i=LstGetSelection(lst);
		if (i!=noListSelection) {
			SerialSpeed=portspeedlist[i];
			if (SerialSpeed==1) {
				SerialSpeed=115200;
				isBlueTooth=true;
				isUSB=false;
			} else if (SerialSpeed==-1) {
				SerialSpeed=115200;
				isBlueTooth=false;
				isUSB=true;
			} else {
				isBlueTooth=false;
				isUSB=false;
			}
			retval=true;
		}
	}		
	FrmDeleteForm(portform);
	FrmSetActiveForm(mainform);
	return retval;
}

// ******************** HandleAbortButton *********************
// ******* React to buttons, mainly to abort transfer *********
// ************************************************************

Boolean HandleAbortButton(EventType *event,BOOL* pfAborted)
{
	switch(event->eType) {
	case ctlSelectEvent:
		switch(event->data.ctlEnter.controlID) {
		case ID_STARTSTOP:
			*pfAborted=true;
			return true;
		case ID_CONFIGURE:
			if (waitingforcommand) {   // idle waiting for next command -> abort+close
				EvtAddEventToQueue(event);
				*pfAborted=true;
			} else {
				SndPlaySystemSound(sndWarning);
				ShowStatus(disconnectstr);
				AddStringToHistory(disconnectstr);
			}
			return true;
		}
	default:
		break;
	}
	return false;
}

// ********************** MessageLoop *************************
// ********* Handle events while program is busy **************
// ************************************************************

BOOL MessageLoop()		// true means abort
{
	EventType aEvent;
	BOOL fAborted;

	fAborted=false;

	EvtGetEvent(&aEvent, 0);

	if (aEvent.eType == nilEvent)
		return false;
	if (aEvent.eType == appStopEvent) {
		if (waitingforcommand) {   // idle waiting for next command -> abort+close
			EvtAddEventToQueue(&aEvent);
			return true;
		} else {
			SndPlaySystemSound(sndWarning);
			ShowStatus(disconnectstr);
			AddStringToHistory(disconnectstr);
			return false;
		}
	}
	// Give the system a chance to handle the event.
	if (!SysHandleEvent(&aEvent))
		// Try to handle the event ourselves.
		if (!HandleAbortButton(&aEvent,&fAborted))
		// Let the form provide default handling of the event.
			FrmHandleEvent(FrmGetActiveForm(), &aEvent);
	return fAborted;
}

// ****************** PalmSysTicksPerSecond********************
// *** Get number of ticks per second, palm-device specific ***
// ************************************************************

UInt32 GloblSysTicks=0;

static UInt32 PalmSysTicksPerSecond()
{
	if (GloblSysTicks==0) {
		if (sysGetROMVerMajor(romversion) < 2)
			GloblSysTicks=sysTicksPerSecond;	// function not defined on version 1!
		else
			GloblSysTicks=SysTicksPerSecond();
	}
	return GloblSysTicks;
}

// ******************** ListDrawFunction **********************
// ** Callback function to draw items in ownerdrawn listbox ***
// ************************************************************

void ListDrawFunction(Int16 itemNum,RectangleType *bounds,Char **itemsText)
{
	if (itemNum<historycount && historylist[itemNum]) {
		WinDrawChars(historylist[itemNum],StrLen(historylist[itemNum]),bounds->topLeft.x,bounds->topLeft.y);
	}
}

// ****************** MainReceiveFunction *********************
// *** Main loop to receive and handle commands from client ***
// ************************************************************

void MainReceiveFunction(void)
{
	Err err;
	UInt8 ch;
	FileRef FileHandle;
	char FileName[260];
	UInt32 volIterator;
	UInt32 vfsMgrVersion;
	UInt32 attributes,fileDate;
	UInt32 restartat,seekto;
	UInt16 receivebuffersize;
	char* receivebuffer=NULL;
	char* block;
	char* p;
	char cmd;
	BOOL overwritefile,isadir;

	if (looprunning)		// avoid reentrancy
		return;

	if ((isHandspring || isSony) && isUSB)
		MessageBox("Please make sure that HotSync isn't running on the PC! It could cause a blue screen crash.",MB_ICONINFORMATION);

	maxblockpart=defmaxblockpart;
	restartat=0;
	overwritefile=false;
	block=NULL;

	VolRefNumCount=0;

	// Make sure that we have access to the virtual file system!
	// Currently only a single SD card slot is supported
	err = FtrGet(sysFileCVFSMgr,vfsFtrIDVersion, &vfsMgrVersion); 
	if (err) {
		// VFS Manager not installed 
		AddStringToHistory("No card file system installed!");
	} else {
		volIterator=vfsIteratorStart;
		VolRefNumCount=0;
		do {
			err=VFSVolumeEnumerate(&VolRefNumList[VolRefNumCount], &volIterator);
			if (err==errNone)
				VolRefNumCount++;
		} while (err==errNone && VolRefNumCount<10);

		if (err!=errNone && err!=vfsIteratorStop && VolRefNumCount<1) {
			AddStringToHistory("No card found!");
			VolRefNumCount=0;
		} else {
			StrIToA(FileName,VolRefNumCount);
			StrCat(FileName," volumes found");
			AddStringToHistory(FileName);
		}
	}
	looprunning=true;

	// Check whether the port is already open
	if (SerialPortHandle==0) {
		err = SerialOpenPort(&SerialPortHandle,SerialSpeed);
		if (err) {
			MessageBox("Could not open com port!",MB_ICONSTOP);
			looprunning=false;
		}
	}

	SetButtonText("Abort+Disconnect");
	AddStringToHistory("Ready");
	{
		SerialClearErr(SerialPortHandle);

		// Try to allocate big buffer - if not possible, use internal(slower)
		receivebuffersize=64512+32;
		receivebuffer =(char*)MemPtrNew(receivebuffersize);
		if (receivebuffer)
			SerialSetReceiveBuffer(SerialPortHandle,receivebuffer,receivebuffersize);
		else
			maxblockpart=256;

		maxblocksize=defmaxblocksize;
		block =(char*)MemPtrNew(maxblocksize);
		if (!block) {
			// device has very limited heap
			// first, make room by freeing the receive buffer
			if (receivebuffer) {
				SerialSetReceiveBuffer(SerialPortHandle,NULL,0);	// restore default buffer
				MemPtrFree(receivebuffer);
				maxblockpart=256;
			}
			// try again to alloc block
			block =(char*)MemPtrNew(maxblocksize);
			if (!block) {
				// if not USB, 1k buffer is sufficient
				maxblocksize=1024;
				block =(char*)MemPtrNew(maxblocksize);
				if (block) {
					AddStringToHistory("USBI (small blocks) or serial on PC!");
					AddStringToHistory("Using 1k reduced buffer size, use");
				}
			}
			if (block) {
				// try to allocate also a 1k receive buffer
				receivebuffersize=1024+32;
				receivebuffer =(char*)MemPtrNew(receivebuffersize);
				if (receivebuffer) {
					SerialSetReceiveBuffer(SerialPortHandle,receivebuffer,receivebuffersize);
					maxblockpart=1024;
				}
			} else {
				MessageBox("Not enough memory to allocate buffer!",MB_ICONSTOP);
				looprunning=false;
			}
		}

		// Main command loop starts here!

		while (looprunning) {
			ch=RQ_TRANSFERERR;
			ShowStatus("Ready...");
			ShowName("-");

			err=SerialSendFlush(SerialPortHandle);		// delete any unsent data
			SerialClearErr(SerialPortHandle);

			// Wait for next command from server
			FileName[0]=0;
			waitingforcommand=true;		// user can switch to other program
			cmd=ReceiveNextBlock(FileName,sizeof(FileName)-1,NULL,NULL);
			waitingforcommand=false;

			// Make forward slashes from backslashes!
			p=FileName;
			while (p[0]) {
				if (p[0]=='\\')
				p[0]='/';
				p++;
			}

			switch(cmd) {
			case CMD_CRCERR:
				ShowStatus("CRC error reported!");
				err = SendChar(RQ_CRCERR);
				break;
			case CMD_ABORT:				// user-abort
				looprunning=false;		// closes the loop!
				break;
			case CMD_CONNECT:			// sort of ping to check that the server is responding
				ShowStatus("Connect!");
				ch=RQ_OK;
				err=SendChar(ch);
				AddStringToHistory("Connect!");
				break;
			case CMD_LIST:
			case CMD_GETSIZE:			// client wants to get only file size
			case CMD_GET:				// client tries to get file or dir
				isadir=false;
				FileHandle=NULL;
				if (FileName[0]=='/' && FileName[1]==0) {		// the root
					if (restartat) {
						ShowStatus("Resume not supported!");	// no resume with dirs!
						ch=RQ_TRANSFERERR;
					} else {
						ShowStatus("GET valid...");
						ch=RQ_OK;
					}
					FileHandle=0;
				} else if (FileName[1]=='0' && FileName[2]==':') {	// a database, or database root
					ControlType* romcheckbox;
					romcheckbox=FrmGetObjectPtr(mainform,FrmGetObjectIndex(mainform,ID_ROM));
					if (romcheckbox)
						wantromdatabases=CtlGetValue(romcheckbox)!=0;
					if (StrLen(FileName)>4 && (cmd==CMD_LIST || romversion<0x03000000)) {
						ShowStatus("GET invalid...");
						ch=RQ_TRANSFERERR;			// ExgDBRead function not supported
					} else {
						if (restartat && StrLen(FileName)<=4) {
							ShowStatus("Resume not supported!");
							ch=RQ_TRANSFERERR;		// no resume with dirs!
						} else {
							ShowStatus("GET valid...");
							ch=RQ_OK;
						}
					}
					FileHandle=0;
				} else if (VolRefNumCount) {
					VolRefNumCurrent=FileName[1]-'1';
					if (VolRefNumCurrent>=0 && VolRefNumCurrent<VolRefNumCount)
						VolRefNumCurrent=VolRefNumList[VolRefNumCurrent];
					if (FileName[3]==0)
						StrCat(FileName,"/");  // the root
					err=VFSFileOpen(VolRefNumCurrent, FileName+3, vfsModeRead, &FileHandle);
					// make sure we get the right object!
					if (err==errNone) {
						err=VFSFileGetAttributes(FileHandle, &attributes);
						isadir=(attributes & vfsFileAttrDirectory)!=0;
						if (isadir!=(cmd==CMD_LIST)) {
							AddStringToHistory("Not a directory!");
							VFSFileClose(FileHandle);
							FileHandle=NULL;
							err=1;
						}
					}
					if (err!=errNone) {
						ShowStatus("GET invalid...");
						ch=RQ_FAILED;
					} else {
						ShowStatus("GET valid...");
						ch=RQ_OK;
					}
				}
				err = SendChar(ch);
				if (err==errNone && ch==RQ_OK) {
					ShowName(FileName);
					err=SerialSendFile(FileHandle, FileName, block, restartat, cmd==CMD_GETSIZE,isadir);
					if (err)
						AddCommandToHistory(cmd==CMD_GETSIZE ? "!SIZE" : "!GET",FileName);
					else
						AddCommandToHistory(cmd==CMD_GETSIZE ? "SIZE" : "GET",FileName);
				} else if (err) {
					StrCopy(FileName,"Error ");
					StrIToA(FileName+StrLen(FileName),err);
					StrCat(FileName," in send");
					AddCommandToHistory("!GET",FileName);
				}
				if (FileHandle)
					VFSFileClose(FileHandle);

				FileHandle=NULL;
				break;
			case CMD_SEND:		// client sends file or dir
				if (FileName[2]!=':' || (FileName[1]=='0' && restartat)) {	// copy to root
					SendChar(RQ_TRANSFERERR);
					break;
				}
				FileHandle=NULL;
				if (FileName[1]=='0') {
					// a database!
					int l;

					l=StrLen(FileName+4);
					if (l>4)
						FileName[l]=0;	// cut off extension .pdb or .prc

					ch=RQ_OK;
					if (!overwritefile) {   // check whether database exists
						LocalID databaseid;
						databaseid=DmFindDatabase(0,FileName+4);
						if (databaseid)
							ch=RQ_FAILED;
					}
					if (ch==RQ_OK)
						ShowStatus("PUT database valid...");
				} else if (VolRefNumCount) {
					VolRefNumCurrent=FileName[1]-'1';
					if (VolRefNumCurrent>=0 && VolRefNumCurrent<VolRefNumCount)
						VolRefNumCurrent=VolRefNumList[VolRefNumCurrent];

					seekto=0xFFFFFFFF;
					if (restartat || overwritefile)
						err=1;
					else
						err=VFSFileOpen(VolRefNumCurrent, FileName+3, vfsModeRead, &FileHandle);
					if (err==errNone) {
						VFSFileClose(FileHandle);
						FileHandle=0;
						ch=RQ_FAILED;
					} else {
						if (restartat)
							err=VFSFileOpen(VolRefNumCurrent, FileName+3, vfsModeWrite | vfsModeExclusive
										, &FileHandle);
						else
							err=VFSFileOpen(VolRefNumCurrent, FileName+3, vfsModeWrite | vfsModeExclusive
										| vfsModeCreate | vfsModeTruncate, &FileHandle);

						if (err==0 && restartat) {
							err=VFSFileSeek(FileHandle,vfsOriginBeginning,restartat);
							if (err && err!=vfsErrFileEOF)
								AddStringToHistory("Seek for restart failed!");
							else
								err=0;
						}
					}

					if (err!=errNone) {
						ShowStatus("PUT invalid...");
						ch=RQ_TRANSFERERR;
					} else {
						ShowStatus("PUT valid...");
						ch=RQ_OK;
					}
				}
				SerialClearErr(SerialPortHandle);

				err = SendChar(ch);
				if (err==errNone && ch==RQ_OK) {
					ShowName(FileName);
					err=SerialReceiveFile(FileHandle, FileName, block, restartat,&seekto);
					if (err)
						AddCommandToHistory("!PUT",FileName);
					else
						AddCommandToHistory("PUT",FileName);
				} else if (err) {
					StrCopy(FileName,"Error ");
					StrIToA(FileName+StrLen(FileName),err);
					StrCat(FileName," in recv");
					AddCommandToHistory("!PUT",FileName);
				}
				if (FileHandle) {
					UInt32 WrittenSize;
					// Resize if file has different size than allocated
					if (0==VFSFileTell(FileHandle,&WrittenSize)) {
						if (WrittenSize<seekto)
							VFSFileResize(FileHandle,WrittenSize);
					}
					VFSFileClose(FileHandle);
					if (seekto==0xFFFFFFFF || (seekto>0 && WrittenSize==0))  //0 bytes sent for non-zero file -> delete
						VFSFileDelete(VolRefNumCurrent, FileName+3);
				}
				FileHandle=NULL;
				break;
			case CMD_DELETE:		// delete file or dir
				ShowStatus("Delete file");
				ShowName(FileName);
				if (VolRefNumCount && FileName[0]=='/' && FileName[1]>='1' && FileName[2]==':') {
					VolRefNumCurrent=FileName[1]-'1';
					if (VolRefNumCurrent>=0 && VolRefNumCurrent<VolRefNumCount)
						VolRefNumCurrent=VolRefNumList[VolRefNumCurrent];

					err=VFSFileDelete(VolRefNumCurrent, FileName+3);
					if (!err) {
						AddCommandToHistory("DEL",FileName);
						ch=RQ_OK;
					} else {
						AddCommandToHistory("!DEL",FileName);
						ch=RQ_FAILED;
					}
					err = SendChar(ch);
				} else {
					AddStringToHistory("DEL on palm card not supported!");
					ch=RQ_FAILED;
					err = SendChar(ch);
				}
				break;
			case CMD_MKDIR:		// delete file or dir
				ShowStatus("Create directory");
				if (VolRefNumCount && FileName[0]=='/' && FileName[1]>='1' && FileName[2]==':') {
					VolRefNumCurrent=FileName[1]-'1';
					if (VolRefNumCurrent>=0 && VolRefNumCurrent<VolRefNumCount)
						VolRefNumCurrent=VolRefNumList[VolRefNumCurrent];

					ShowStatus("Create directory");
					ShowName(FileName);
					err=VFSDirCreate(VolRefNumCurrent, FileName+3);
					if (!err) {
						ch=RQ_OK;
						AddCommandToHistory("MKDIR",FileName);
					} else
						AddCommandToHistory("!MKDIR",FileName);
				}
				err = SendChar(ch);
				break;
			case CMD_RENFR:			// renamefrom file or dir
				StrLCopy(renamefromname,FileName,sizeof(renamefromname)-1);
				AddCommandToHistory("RNFR",FileName);
				ch=RQ_OK;
				err = SendChar(ch);
				break;
			case CMD_RENTO:			// renameto file or dir
				ShowStatus("Rename file");
				if (VolRefNumCount && renamefromname[0] && FileName[1]>='1' && renamefromname[2]==':') {
						  char *p1,*p2;
					VolRefNumCurrent=FileName[1]-'1';
					if (VolRefNumCurrent>=0 && VolRefNumCurrent<VolRefNumCount)
						VolRefNumCurrent=VolRefNumList[VolRefNumCurrent];

					p1=StrRChr(renamefromname,'/');
					p2=StrRChr(FileName,'/');
					if (p1 && p2 && p1-renamefromname==p2-FileName) {
						if (0==VFSFileRename(VolRefNumCurrent,renamefromname+3,p2+1)) // only same dir!
							ch=RQ_OK;
					} else
						ch=RQ_FAILED;
						
				} else
					ch=RQ_FAILED;
				if (ch==RQ_OK)
					AddCommandToHistory("RNTO",FileName);
				else
					AddCommandToHistory("!RNTO",FileName);
				err = SendChar(ch);
				break;
			case CMD_SETATTR:
				p=StrChr(FileName,'\t');
				err=1;
				if (VolRefNumCount && p && FileName[0]=='/' && FileName[1]>='1' && FileName[2]==':') {
					p[0]=0;
					p++;
					attributes=StrAToI(p);
					ShowStatus("Change attributes");
					ShowName(FileName);
					VolRefNumCurrent=FileName[1]-'1';
					if (VolRefNumCurrent>=0 && VolRefNumCurrent<VolRefNumCount)
						VolRefNumCurrent=VolRefNumList[VolRefNumCurrent];

					err=VFSFileOpen(VolRefNumCurrent, FileName+3, vfsModeWrite | vfsModeExclusive, &FileHandle);
					if (!err) {
						err=VFSFileSetAttributes(FileHandle,attributes);
						VFSFileClose(FileHandle);
					}
				}
				if (!err) {
					ch=RQ_OK;
					AddCommandToHistory("ATTR",FileName);
				} else
					AddCommandToHistory("!ATTR",FileName);

				err = SendChar(ch);
				break;
			case CMD_SETTIME:
				p=StrChr(FileName,'\t');
				if (VolRefNumCount && p && FileName[0]=='/' && FileName[1]>='1' && FileName[2]==':') {
					p[0]=0;
					p++;
					AddCommandToHistory("TIME:",p);
					fileDate=ConvertIsoDateToDateTime(p);
					if (fileDate) {
						ShowStatus("Change date/time");
						ShowName(FileName);
						VolRefNumCurrent=FileName[1]-'1';
						if (VolRefNumCurrent>=0 && VolRefNumCurrent<VolRefNumCount)
							VolRefNumCurrent=VolRefNumList[VolRefNumCurrent];

						err=VFSFileOpen(VolRefNumCurrent, FileName+3, vfsModeWrite | vfsModeExclusive, &FileHandle);
						if (!err) {
							err=VFSFileSetDate(FileHandle,vfsFileDateModified,fileDate);
							if (!err) {
								ch=RQ_OK;
								AddCommandToHistory("TIME",FileName);
							} else
								AddCommandToHistory("!TIME",FileName);
							VFSFileClose(FileHandle);
						}
					}
				}
				err = SendChar(ch);
				break;
			case CMD_RESTART:
				restartat=StrAToI(FileName);
				if (restartat>0)
					AddCommandToHistory("RESTART at",FileName);
				ch=RQ_OK;
				err = SendChar(ch);
				break;
			case CMD_OVERWRITE:
				overwritefile=1;
				AddStringToHistory("Overwrite enabled");
				ch=RQ_OK;
				err = SendChar(ch);
				break;
			default:
				ch=RQ_FAILED;
				err = SendChar(ch);
			}
			if (cmd!=CMD_RESTART && cmd!=CMD_CRCERR)	// only use restart for next command!
				restartat=0;
			if (cmd!=CMD_OVERWRITE && cmd!=CMD_CRCERR)	// only use overwrite for next command!
				overwritefile=0;
		}
	}
	// Change the status text.

	ShowStatus("Finished");
	SetButtonText("Connect");

	if (receivebuffer) {
		SerialSetReceiveBuffer(SerialPortHandle,NULL,0);	// restore default buffer
		MemPtrFree(receivebuffer);
	}
	if (block)
		MemPtrFree((MemPtr)block);

	SerialClose(SerialPortHandle);		// close any open port before applying new settings
	SerialPortHandle=NULL;
	looprunning=false;

	if (continuesleepneeded) {
		// After you defer sleep, your code is free to finish what it was doing.
		// When it is finished, you must allow the system to continue with the
		// sleep event. To do so, create a keyDownEvent with the
		// resumeSleepChr and the command key bit set (to signal that the
		// character is virtual) and add it to the event queue.
		EventType event;
		event.eType=keyDownEvent;
		event.data.keyDown.chr=resumeSleepChr;
		event.data.keyDown.keyCode=0;
		event.data.keyDown.modifiers=commandKeyMask;
		EvtAddEventToQueue(&event);
		AddStringToHistory("Disconnecting for sleep mode");
		
		continuesleepneeded=false;
	}
}

BOOL IsCommand(int cmd)
{
 return(cmd>='A' && cmd<='Z');  // commands are uppercase characters
}

// ******************** TimedReceiveChar **********************
// ************************************************************
// Receive from Comm port, allow abort
// timeout_msec: timeout in milliseconds, -1 for infinite
// ************************************************************

Err TimedReceiveChar(char* ch,Int32 timeout_msec)
{
	UInt32 tries;
	UInt32 maxtries=timeout_msec/250;
	Err err;
	if (timeout_msec<0)
		maxtries=-1;
	tries=0;

	do {
		EvtResetAutoOffTimer();
		SerialReceive(SerialPortHandle,ch,1,PalmSysTicksPerSecond()/4,&err);
		if (err)
			SerialClearErr(SerialPortHandle);
		if (MessageLoop()) {
			ShowStatus("Aborting 1...");
			return err_abort;
		}
		tries++;
	} while (err==serErrTimeOut &&(timeout_msec<0 || tries<=maxtries));

	if (err)
		return err_recv;
	return 0;
}

// ******************** TimedReceiveBytes *********************
// ************************************************************
// Receive exactly numbytes bytes
// Receive from Comm port, allow abort
// timeout_msec: timeout in milliseconds, -1 for infinite
// ************************************************************

Err TimedReceiveBytes(char* recvbuf,UInt32* numbytes,Int32 timeout_msec)
{
	UInt32 tries;
	UInt32 maxtries=timeout_msec/250;
	Int32 numb=(Int32)*numbytes;
	Err err;
	tries=0;

	// Wait for first numbytes bytes of command
	do {
		EvtResetAutoOffTimer();
		err=SerialReceiveWait(SerialPortHandle,numb,PalmSysTicksPerSecond()/4);
		if (err) {
			*numbytes=0;
			SerialClearErr(SerialPortHandle);
		}
		if (MessageLoop()) {
			ShowStatus("Aborting 2...");
			return err_abort;
		}
		tries++;
	} while (err==serErrTimeOut &&(timeout_msec<0 || tries<=maxtries));
	if (!err) {
		SerialReceive(SerialPortHandle,recvbuf,numb,PalmSysTicksPerSecond()/2,&err);
		if (err) {
			*numbytes=0;
			SerialClearErr(SerialPortHandle);
		}
	}
	if (err)
		return err_send;
	return 0;
}

// ************************ SendBytes *************************
// ************************************************************
// ** Send exactly numbytes bytes to com port, allow abort. ***
// Note:
// Unfortunately SrmSend doesn't seem to work correctly when a
// CTS timeout occurs, returning an invalid value for bytes
// sent, so simply set a high value for CTS timeout
// ************************************************************

Err SendBytes(char* sendbuf,Int32 numbytes)
{
	Err err;

	err=0;
	// Try to send whole block, react to CTS timeout
	SerialSend(SerialPortHandle,sendbuf,numbytes,&err);
	if (err) {
		SerialClearErr(SerialPortHandle);
		if (MessageLoop()) {
			AddStringToHistory("Aborting 3...");
			ShowStatus("Aborting 3...");
			return err_abort;
		}
	}
	EvtResetAutoOffTimer();
	return err;
}

// ************************ SendChar **************************
// ** Send a single character to the port, clear comm errors **
// ************************************************************

Err SendChar(char ch)
{
	return SendBytes(&ch,1);
}

// ********************* TimedSendChar ************************
// ****** Send 1 byte to the com port, wait timeout_msec ******
// Note: timeout is ignored here!
// ************************************************************

Err TimedSendChar(char ch,Int32 timeout_msec)
{
	return SendChar(ch);
}

// ********************* TimedSendBytes ***********************
// ** Send numbytes bytes to the com port, wait timeout_msec **
// Note: timeout is ignored here!
// ************************************************************

Err TimedSendBytes(char* sendbuf,Int32 numbytes,Int32 timeout_msec)
{
	return SendBytes(sendbuf,numbytes);
}

// ************************ Connected *************************
// *********** Check whether the port is open or not **********
// ************************************************************

BOOL Connected()
{
	return SerialPortHandle!=0;
}

// ********************** SendNextBlock ***********************
// ****** Send next block when sending a file to the PC *******
// ************************************************************

int SendNextBlock(char request,char* databuf,UInt32 datalen,UInt8 sequencenr)
{
	Int32 l;
	UInt16 crc;
	Err err;
	char headerbuf[8];
	char crcbuf[2];
	char bytereceived;

	if (!Connected())
		return RQ_TRANSFERERR;
	
	headerbuf[0]='F';
	headerbuf[1]='T';
	if (request==CMD_BODY)
		headerbuf[2]=(char)sequencenr;
	else
		headerbuf[2]='H';
	headerbuf[3]=request;
	l=8+datalen;
	if (datalen)
		l+=2;	// the crc!
	headerbuf[4]=(l >> 8) & 255;
	headerbuf[5]=l & 255;
	crc = Crc16CalcBlock(headerbuf, 6, 0);
	headerbuf[6]=(char)(crc >> 8);
	headerbuf[7]=(char)crc;

	// Send header
	err=TimedSendBytes(headerbuf,8,5000);
	if (err) {
		ShowStatus("Header Write error!");
		  if (err==err_abort)
			return RQ_LOCALABORT;
		else
			return RQ_LOCALTIMEOUT;
	}

	if (datalen>0) {
		// Important: calculate CRC before sending data!
		// Reason: Receiving the final CRC is time-critical
		// because if 1-2 bytes are lost, filepc2pda has to react quickly
		crc = Crc16CalcBlock(&headerbuf[6], 2, crc);
		crc = Crc16CalcBlock(databuf, datalen, crc);
		crcbuf[0]=(char)(crc >> 8);
		crcbuf[1]=(char)crc;

		// Send body
		err=TimedSendBytes(databuf,datalen,5000);
		if (err) {
			ShowStatus("Body Write error!");
			  if (err==err_abort)
				return RQ_LOCALABORT;
			else
				return RQ_LOCALTIMEOUT;
		}

		// Send crc
		err=TimedSendBytes(crcbuf,2,5000);
		if (err) {
			ShowStatus("CRC Write error!");
			  if (err==err_abort)
				return RQ_LOCALABORT;
			else
				return RQ_LOCALTIMEOUT;
		}
	}

	err=TimedReceiveChar(&bytereceived,5000);
	if (err) {
		  if (err==err_abort) {
			return RQ_LOCALABORT;
		} else
			return RQ_LOCALTIMEOUT;
	}
	while (!(bytereceived==RQ_OK || bytereceived==RQ_CRCERR || bytereceived==RQ_FAILED || bytereceived==RQ_TRANSFERERR)) {
		err=TimedReceiveChar(&bytereceived,100); // delete any bytes before the actual reply!
		if (err==err_abort) {
			return RQ_LOCALTIMEOUT;
		}
	}
	if (bytereceived==RQ_OK || bytereceived==RQ_CRCERR || bytereceived==RQ_FAILED)
		return bytereceived;
	return RQ_TRANSFERERR;
}

// ****************** SendNextBlockRetry10 ********************
// * SendNextBlock with retry in case of timeout or CRC error *
// ************************************************************

int SendNextBlockRetry10(char request,char* databuf,UInt32 datalen,UInt8 sequencenr)
{
	Err err;
	Int16 retries;

	retries=0;
	do {
		err=SendNextBlock(request,databuf,datalen,sequencenr);
		if (err==RQ_CRCERR) {
			SerialSendFlush(SerialPortHandle);		// delete any unsent data
			AddStringToHistory("GET CRC error");
		}
		if (err==RQ_LOCALTIMEOUT) {
			AddStringToHistory("Timeout!");
			SerialSendFlush(SerialPortHandle);		// delete any unsent data
			SerialClearErr(SerialPortHandle);
			Sleep(4000);
		}
	} while ((err==RQ_CRCERR || err==RQ_LOCALTIMEOUT) && retries++<10);
	return err;
}

// ******************** ReceiveNextBlock **********************
// ** Receive next command or data block. Returns block type **
// ** or an error(CMD_xxx constants) *************************
// ************************************************************

// returns CMD_ABORT, CMD_CRCERR, CMD_RECVERR, CMD_TIMEOUT or a valid command
char ReceiveNextBlock(char* recvbody,UInt32 maxbodysize,UInt32* datareceived,UInt8* sequencenr)
{
	Err err;
	UInt8 headerbuf[10];
	UInt8 crcbuf[2];
	UInt32 packetlen;
	UInt32 blockread;
	UInt32 fullblocksize;
	UInt32 thisblockpart;
	UInt32 receive;
	UInt16 crc,headcrc,fullcrc;
	UInt8* p;

	if (!Connected())
		return CMD_RECVERR;
	
	headerbuf[0]=0;
	headerbuf[8]=0;
	headerbuf[9]=0;
	// Wait for first 8 bytes of command

	receive=8;
	err=TimedReceiveBytes((char*)&headerbuf,&receive,-1);	// wait indefinitely or until user clicks cancel
	if (err==err_abort)
		return CMD_ABORT;

	p=headerbuf;
	while (p<headerbuf+6 && !(p[0]=='F' && p[1]=='T' &&(p[2]=='H' || p[3]==CMD_BODY)))
		p++;
	if (p!=headerbuf && p<headerbuf+6) {	// found header!
		ShowStatus("Try to fix header!");
		memmove(&headerbuf, p, 8-(p-headerbuf));
		p=headerbuf+8-(p-headerbuf);
		receive=8-(p-headerbuf);
		err=TimedReceiveBytes((char*)p,&receive,-1);	// wait indefinitely or until user clicks cancel
		if (err==err_abort)
			return CMD_ABORT;
	}
	if (sequencenr)
		*sequencenr=(UInt8)headerbuf[2];
	packetlen=(UInt32)256*headerbuf[4]+headerbuf[5];

	crc = Crc16CalcBlock((char*)headerbuf, 6, 0);
	headcrc =(UInt16)256*headerbuf[6]+headerbuf[7];

	// Three cases:
	// 1. Completely received header with correct CRC -> get body
	// 2. Completely received header with errors -> flush input
	// 3. Receive error(not enough bytes) -> return error

	if (!err && headerbuf[0]=='F' && headerbuf[1]=='T' &&(headerbuf[2]=='H' || headerbuf[3]==CMD_BODY) &&
		(crc==headcrc) &&
		packetlen>=8 && packetlen<=maxbodysize+10) {
		// Case 1:
		// Wait for remaining packetlen-8 bytes of command
		blockread=0;
		if (packetlen<10)
			fullblocksize=0;
		else
			fullblocksize=packetlen-10;	//minus header and CRC!
		if (datareceived)
			*datareceived=fullblocksize;
		while (blockread<fullblocksize) {
			// Receive block in 256 byte chunks
			thisblockpart=fullblocksize-blockread;
			if (thisblockpart>maxblockpart)	// only works with larger receive buffer above!
				thisblockpart=maxblockpart;
			err=TimedReceiveBytes(recvbody+blockread,&thisblockpart,5000);	// wait up to 5 seconds for the rest
			if (err==err_abort) {
				return CMD_ABORT;
			} else if (err) {
				ShowStatus("Transfer timeout...");
				if (!Connected())
					return CMD_RECVERR;
				else
					return CMD_TIMEOUT;
			}
			blockread+=thisblockpart;
		} // end while
		// wait for crc of whole block
		if (fullblocksize) {
			receive=2;
			err=TimedReceiveBytes((char*)&crcbuf,&receive,1000);	// wait only 1 second for CRC. Reason: if 1-2 bytes are lost
			fullcrc=(UInt16)256*crcbuf[0]+crcbuf[1];
			if (err==err_abort)
				return CMD_ABORT;
			crc = Crc16CalcBlock((char*)headerbuf+6, 2, crc);
			crc = Crc16CalcBlock(recvbody, fullblocksize, crc);
			if (crc!=fullcrc) {
				ShowStatus("BAD CRC of body");
				return CMD_CRCERR;
			}
		}
	} else if (err==0) {	// case 2: data recieved, but bad
		// to do: get rid of any extra bytes in front of the header!
		if (!(headerbuf[0]=='F' && headerbuf[1]=='T'))
			ShowStatus("Bad header!");
		else if (crc!=headcrc)
			ShowStatus("Bad CRC");
		else ShowStatus("Header bad");
		SerialClearReceiveBuffer();
		err=CMD_CRCERR;
	} else {	// case 3: data not received
		err=CMD_TIMEOUT;
		ShowStatus("Timeout!");
		recvbody[0]=0;
	}
	if (!err) {
		return headerbuf[3];
	} else
		return err;
}

Err ReceiveNextDataBlockRetry10(char* recvbody,UInt32 maxbodysize,UInt32* datareceived,UInt8* expectedsequencenr)
{
	Err err;
	int retries;
	UInt8 sequencenr;
	// Receive a whole block
	retries=0;
	do {
		SerialClearErr(SerialPortHandle);
		err=ReceiveNextBlock(recvbody,maxbodysize,datareceived,&sequencenr);
		if (err==CMD_BODY && *expectedsequencenr!=sequencenr) {
			AddStringToHistory("Wrong sequence number!");
			err=CMD_CRCERR;	// wrong block received!
		}
		if (IsCommand(err)) {	 // upload was aborted, this is already a new command!
			AddStringToHistory("Download interrupted!");
			SendChar(RQ_CRCERR); // requests re-send
			return CMD_CRCERR;	// wrong block received!
		}
		if (err==CMD_CRCERR) {
			ShowStatus("CRC-Error");
			if (SerialClearReceiveBuffer()==err_abort)
				return CMD_ABORT;
			AddStringToHistory("PUT CRC error");
			TimedSendChar(RQ_CRCERR,5000);
		}
		if (err==CMD_TIMEOUT) {
			ShowStatus("Timeout!");
			AddStringToHistory("PUT Timeout!");
			Sleep(1000);
			if (SerialClearReceiveBuffer()==err_abort)
				return err_abort;
			SerialSendFlush(SerialPortHandle);		// delete any unsent data
			SerialClearErr(SerialPortHandle);
			SendChar(RQ_CRCERR);
			Sleep(2000);
		}
	} while ((err==CMD_CRCERR || err==CMD_TIMEOUT) && retries++<10);

	if (err==CMD_BODY) {
		if ((*expectedsequencenr)==255)
			*expectedsequencenr=0;
		else
			(*expectedsequencenr)++;
	}
	return err;
}

Boolean ExgDeleteProc(const char* nameP,UInt16 version,UInt16 cardNo,LocalID dbID,void*userDataP)
{
	AddCommandToHistory("Overwrite database: ",(char*)nameP);
	return 0==DmDeleteDatabase(cardNo,dbID);
}

// ********************** ExgReadProc *************************
// ** Callback function to send a database from serial to RAM *
// ************************************************************

Err ExgReadProc(void* dataP, UInt32* sizeP, void* userDataP)
{
	Err err;
	pDBWR puserblock;
	char* pdataout;
	UInt32 sizedatarequested;
	UInt32 bytestocopy,blockbytecontent,offset,datareceived,datacopied;
	UInt8 sequencenrexpected;
	char* block;

	pdataout=(char*)dataP;
	sizedatarequested=*sizeP;

	puserblock=(pDBWR)userDataP;
	block=puserblock->block;
	blockbytecontent=puserblock->blockbytecontent;
	sequencenrexpected=puserblock->sequencenr;
	offset=puserblock->offset;

	datacopied=0;
	
	// loop while there is data requested AND available!
	while (sizedatarequested && (!puserblock->eofreached || blockbytecontent)) {
		// only request new data if no more data available
		if (blockbytecontent==0 && !puserblock->eofreached) {
			err=ReceiveNextDataBlockRetry10(block,maxblocksize,&datareceived,&sequencenrexpected);
			if (err==CMD_ABORT)
				return err_abort;
			if (err==CMD_RECVERR || err==CMD_CRCERR || err==CMD_TIMEOUT)
				return err_recv;
			if (err==CMD_EOF) {
				AddStringToHistory("Last block received!");
				puserblock->eofreached=true;
				SendChar(RQ_OK);
			} else if (err==CMD_BODY) {
				blockbytecontent=datareceived;
				SendChar(RQ_OK);
			} else
				SendChar(RQ_TRANSFERERR);
		}
		// now, copy as much data as possible from received block to request block	
		if (blockbytecontent) {
			bytestocopy=blockbytecontent;
			if (bytestocopy>sizedatarequested)
				bytestocopy=sizedatarequested;
			memmove(pdataout,block,bytestocopy);

			// if not all data from block copied, move remaining to beginning of block
			if (bytestocopy<blockbytecontent)
				memmove(block,block+bytestocopy,blockbytecontent-bytestocopy);
			pdataout+=bytestocopy;
			datacopied+=bytestocopy;
			blockbytecontent-=bytestocopy;
			sizedatarequested-=bytestocopy;
			offset+=bytestocopy;

			if (puserblock->FileSize!=0xFFFFFFFF)
				ShowPercentDone(offset, puserblock->FileSize);
			else {
				StrCat(puserblock->progress,"#");
				if (StrLen(puserblock->progress)>10)
					puserblock->progress[0]=0;
				ShowStatus(puserblock->progress);
			}
		}
	}
	*sizeP=datacopied;
	puserblock->blockbytecontent=blockbytecontent;
	puserblock->sequencenr=sequencenrexpected;
	puserblock->offset=offset;
	return 0;
}

// ******************** SerialReceiveFile *********************
// * Receive a file from the PC via the serial port. Uses the *
// * same type of blocks as for commands, of type CMD_BODY ****
// ************************************************************

Err SerialReceiveFile(FileRef FileHandle,char* FileName,char* block,UInt32 restartat,UInt32* seekto)
{
	Err err;
	UInt32 FileSize;
	UInt32 fileDate;
	UInt32 FileRead;
	UInt32 NumBytesWritten;
	UInt32 DataReceived;
	UInt32 blockoffset;
	UInt8 expectedsequencenr;

	// Signal to PC that we are ready to receive data
	err = SendChar(RQ_READY);

	if (err)
		return err;

	*seekto=0xFFFFFFFF;       // never seeked!
	FileSize=0;
	fileDate=0;
	FileRead=restartat;
	blockoffset=0;
	expectedsequencenr=0;

	ShowStatus("File receive...");

	while (true) {
		EvtResetAutoOffTimer();	//avoid sleep mode

		err=ReceiveNextDataBlockRetry10(block+blockoffset,maxblocksize-blockoffset,&DataReceived,&expectedsequencenr);

		if (err==CMD_ABORT)
			return err_abort;
		if (err==CMD_RECVERR || err==CMD_CRCERR || err==CMD_TIMEOUT)
			return err_recv;

		if (err==CMD_EOF) {
			ShowStatus("Write last block");
			SendChar(RQ_OK);
			// Write to file if there is any data to write left
			if (blockoffset>0) {
				err=VFSFileWrite(FileHandle,blockoffset,block,&NumBytesWritten);
				if (err)
					return err_file;
			}
			if (fileDate)
				VFSFileSetDate(FileHandle,vfsFileDateModified,fileDate);
			break;
		}
		if (err==CMD_SIZE) {	// size block
			char* pdate;
			pdate=StrChr(block,'\t');
			if (pdate) {
				pdate[0]=0;
				fileDate=ConvertIsoDateToDateTime(++pdate);
			}
			FileSize = StrAToI((char*) block);

			*seekto=FileSize;
			if (FileSize>64512 && FileName[1]!='0') {    // only faster for big files!
				ShowStatus("Set file size...");
				err = SendChar(RQ_WAIT);   // intermediate result - tell client to wait an extended time!
				err=VFSFileResize(FileHandle,FileSize);
				if (err) {
					AddStringToHistory("Not enough space for file!");
					return err_file;
				}
				ShowStatus("Resize done");
			}
			// Send acknowledge for block
			err = SendChar(RQ_OK);
			if (err)
				AddStringToHistory("File resize OK failed!");
		} else if (err==CMD_BODY) {	// body
			UInt32 writesize=DataReceived;
			if (writesize>FileSize-FileRead)
				writesize=FileSize-FileRead;
			FileRead += writesize;
			// Send acknowledge for block
			err = SendChar(RQ_OK);

			if (FileName[1]=='0') {  // copy to internal RAM!
				LocalID dbid;
				Boolean needreset;
				DBWR dbreadrec;
				dbreadrec.block=block;
				dbreadrec.blockbytecontent=FileRead;
				dbreadrec.sequencenr=expectedsequencenr;
				dbreadrec.progress[0]=0;
				dbreadrec.FileSize=FileSize;
				dbreadrec.offset=0;
				dbreadrec.eofreached=false;
				err=ExgDBRead(ExgReadProc, ExgDeleteProc, &dbreadrec, &dbid, 0, &needreset,true);
				if (!dbreadrec.eofreached)
					SendChar(RQ_TRANSFERERR);
				if (err)
					AddStringToHistory("ExgDBRead returned an error");
				return err;
			}
			// If less than 16k received, just increase block offset
			// but only if blockoffset+2*DataReceived<maxblocksize
			if (DataReceived<16384 && blockoffset+2*DataReceived<maxblocksize) {
				blockoffset+=DataReceived;
			} else {
				// Write to file if there is any data to write
				if (writesize+blockoffset>0) {
					err=VFSFileWrite(FileHandle,blockoffset+writesize,block,&NumBytesWritten);
					if (err)
						return err_file;
				}
				blockoffset=0;
			}
			ShowPercentDone(FileRead, FileSize);
		}
		MessageLoop();
	}
	return 0;
}

// ********************** ExgWriteProc ************************
// ** Callback function to send a database from RAM to serial *
// ************************************************************

Err ExgWriteProc(const void* dataP, UInt32* sizeP, void* userDataP)
{
	Err err;
	pDBWR puserblock;
	char* pdataremain;
	UInt32 sizedataremain;
	UInt32 sendblocksize,nextchunksize,blockbytecontent,offset;
	UInt8 sequencenr;
	char* block;

	pdataremain=(char*)dataP;
	sizedataremain=*sizeP;

	puserblock=(pDBWR)userDataP;
	block=puserblock->block;
	blockbytecontent=puserblock->blockbytecontent;
	sequencenr=puserblock->sequencenr;
	offset=puserblock->offset;

	if (isUSB)
		sendblocksize=maxblocksize;
	else
		sendblocksize=1024;

	while (sizedataremain>0) {
		// Discard data when resuming
		if (puserblock->restartdiscard>0) {
			nextchunksize=puserblock->restartdiscard;
			if (nextchunksize>sizedataremain)
				nextchunksize=sizedataremain;
			pdataremain+=nextchunksize;
			sizedataremain-=nextchunksize;
			puserblock->restartdiscard-=nextchunksize;
		}
		// First, move as much data as possible from dataP to block
		nextchunksize=sendblocksize-blockbytecontent;
		if (nextchunksize>sizedataremain)
			nextchunksize=sizedataremain;

		if (nextchunksize>0)
			memmove(block+blockbytecontent,pdataremain,nextchunksize);
		blockbytecontent+=nextchunksize;
		pdataremain+=nextchunksize;
		sizedataremain-=nextchunksize;

		// Now send the data, but ONLY if block is full!!!

		if (blockbytecontent==sendblocksize) {
			err=SendNextBlockRetry10(CMD_BODY,(char*)block,sendblocksize,sequencenr);
			if (err!=RQ_OK) {
				if (err==RQ_LOCALABORT) {
					AddStringToHistory("Aborting database download!");
					return err_abort;
				} else {
					AddStringToHistory("Body block not acknowledged!");
					return err_recv;
				}
			 }

			if (sequencenr==255)
				sequencenr=0;
			else
				sequencenr++;

			blockbytecontent=0;	// everything was sent!

			offset+=sendblocksize;

			if (puserblock->FileSize!=0xFFFFFFFF)
				ShowPercentDone(offset, puserblock->FileSize);
			else {
				StrCat(puserblock->progress,"#");
				if (StrLen(puserblock->progress)>10)
					puserblock->progress[0]=0;
				ShowStatus(puserblock->progress);
			}
		}
	}

	puserblock->blockbytecontent=blockbytecontent;
	puserblock->sequencenr=sequencenr;
	puserblock->offset=offset;
	return 0;
}

// ********************* SerialSendFile ***********************
// ** Send a file or directory to the PC. Supports both the ***
// ** internal RAM and external SD cards                    ***
// ************************************************************

Err SerialSendFile(FileRef FileHandle,char* FileName,char* block, UInt32 restartat, BOOL sizeonly,BOOL isadir)
{
	UInt32 FileSize;
	UInt32 dirIterator;
	char bytereceived;
	unsigned char* sendptr;
	UInt32 blocksize,sendblocksize;
	UInt32 offset;
	UInt32 numBytesRead;
	UInt16 currdatabase,numdatabases;
	UInt8 sequencenr;
	LocalID databaseid;
	Err err;
	BOOL istheroot,isadatabase;
	char progress[12];

	SerialClearErr(SerialPortHandle);

	sequencenr=0;
	progress[0]=0;
	dirIterator=vfsIteratorStart;
	isadatabase=false;
	istheroot=false;
	FileSize=0;
	offset=0;
	currdatabase=0;
	numdatabases=0;
	if (FileHandle==0) {		// Enumerate databases
		FileSize=0xFFFFFFFF;
		if (StrLen(FileName)==1)
			istheroot=true;
		else if (StrLen(FileName)==3 || StrLen(FileName)==4)
			numdatabases=DmNumDatabases(0);
		else {
			UInt32 TotalBytes,numrecs;
			UInt16 attributes;
			LocalID appinfoid,sortinfoid;
			char* ptr;
			int l;

			isadatabase=true;
			l=StrLen(FileName+4);
			if (l>4)
				FileName[l]=0;	// cut off extension .pdb or .prc
			databaseid=DmFindDatabase(0,FileName+4);
			if (databaseid && 0==DmDatabaseInfo(0,databaseid,NULL,&attributes,NULL,NULL,NULL,NULL,NULL,&appinfoid,&sortinfoid,NULL,NULL)) {
				if (0==DmDatabaseSize(0,databaseid,&numrecs,NULL,&TotalBytes)) {
					if (attributes & dmHdrAttrResDB)
						TotalBytes+=2+mainheadersize+numrecs*prcrecheadersize;
					else
						TotalBytes+=2+mainheadersize+numrecs*pdbrecheadersize;
					if (appinfoid) {
						ptr=MemLocalIDToLockedPtr(appinfoid,0);
						if (ptr) {
							TotalBytes+=MemPtrSize(ptr);
							MemPtrUnlock(ptr);
						}
					}
					if (sortinfoid) {
						ptr=MemLocalIDToLockedPtr(sortinfoid,0);
						if (ptr) {
							TotalBytes+=MemPtrSize(ptr);
							MemPtrUnlock(ptr);
						}
					}
					FileSize=TotalBytes;
				}
			}
		}
	} else {
		if (!isadir) {
			err=VFSFileSize(FileHandle,&FileSize);
			if (err!=errNone) {
				AddStringToHistory("Could not access file!");
				return err_file;
			}
			// it's a directory!
			if (restartat) {
				err=VFSFileSeek(FileHandle,vfsOriginBeginning,restartat);
				if (err) {
					AddStringToHistory("Restart failed(invalid seek)!");
					return err_file;
				}
				offset=restartat;
			}
		} else
			FileSize=0xFFFFFFFF;
	}

	// Wait for RQ_READY
	err=TimedReceiveChar(&bytereceived,5000);

	if (err) {		// timeout!
		AddStringToHistory("Initial byte timeout!");
		  if (err==RQ_LOCALABORT)
			return err_abort;
		else
			return err_recv;
	}
	if (bytereceived!=RQ_READY) {
		AddStringToHistory("Initial byte receive error!");
		return err_recv;
	}
	// send header block
	StrIToA(block,FileSize);

	err=SendNextBlockRetry10(CMD_SIZE,block,StrLen(block)+1,0);

	if (err!=RQ_OK) {
		AddStringToHistory("Size block not acknowledged!");
		  if (err==RQ_LOCALABORT)
			return err_abort;
		else
			return err_recv;
	}

	if (sizeonly)
		return 0;

	// send body
	EvtResetAutoOffTimer();

	while (offset < FileSize && dirIterator!=vfsIteratorStop)
	{
		MessageLoop();
		if (istheroot) {
			int i;
			char* p;
			StrCopy(block,"0:\\");
			for (i=0;i<VolRefNumCount;i++) {
				p=block+StrLen(block);
				StrCopy(p,"\r\n1:\\");
				p[2]='1'+i;
			}
			blocksize=StrLen(block);	// no #0 at end!
			dirIterator=vfsIteratorStop;
		} else if (isadatabase) {	// Transmit database via callback
			DBWR dbwriterec;
			dbwriterec.block=block;
			dbwriterec.blockbytecontent=0;
			dbwriterec.sequencenr=0;
			dbwriterec.progress[0]=0;
			dbwriterec.FileSize=FileSize;
			dbwriterec.restartdiscard=restartat;
			dbwriterec.offset=0;
			err=ExgDBWrite(ExgWriteProc, &dbwriterec, FileName+4, 0, 0);

			if (err)
				return err;

			dirIterator=vfsIteratorStop; // done

			// prepare to write any remaining bytes
			blocksize=dbwriterec.blockbytecontent;
			sequencenr=dbwriterec.sequencenr;
			offset=dbwriterec.offset;
			if (blocksize==0)		// no bytes remaining!
				break;
		} else if (FileHandle==0) {			// Enumerate databases
			Char *FileName =MemPtrNew(260);
			block[0]=0;
			blocksize=0;
			if (FileName) {
				char* pEndBlock;
				char* pNextName=block;
				UInt32 TotalBytes,numrecs,modDate,creadate;
				UInt16 attributes;
				LocalID appinfoid,sortinfoid;
				char* ptr;
				block[0]=0;
				blocksize=1024;		//maxblocksize;
				pEndBlock=block+blocksize-100;	// if more files, continue in next block
				while (dirIterator!=vfsIteratorStop && pNextName<pEndBlock) {
					LocalID dbid=DmGetDatabase(0,currdatabase);
					if (dbid &&(wantromdatabases || MemLocalIDKind(dbid)!=memIDPtr)) {
						modDate=0;
						creadate=0;
						if (0==DmDatabaseInfo(0,dbid,FileName,&attributes,NULL,&creadate,&modDate,NULL,NULL,&appinfoid,&sortinfoid,NULL,NULL)) {
							StrCopy(pNextName,FileName);
							pNextName+=StrLen(FileName);
							if (attributes & dmHdrAttrResDB)
								StrCopy(pNextName,".prc");
							else
								StrCopy(pNextName,".pdb");
							pNextName+=4;
							if (0==DmDatabaseSize(0,dbid,&numrecs,NULL,&TotalBytes)) {
								pNextName[0]='\t';
								if (attributes & dmHdrAttrResDB)
									TotalBytes+=2+mainheadersize+numrecs*prcrecheadersize;
								else
									TotalBytes+=2+mainheadersize+numrecs*pdbrecheadersize;
								if (appinfoid) {
									ptr=MemLocalIDToLockedPtr(appinfoid,0);
									if (ptr) {
										TotalBytes+=MemPtrSize(ptr);
										MemPtrUnlock(ptr);
									}
								}
								if (sortinfoid) {
									ptr=MemLocalIDToLockedPtr(sortinfoid,0);
									if (ptr) {
										TotalBytes+=MemPtrSize(ptr);
										MemPtrUnlock(ptr);
									}
								}
								StrIToA(pNextName+1,TotalBytes);
								pNextName+=StrLen(pNextName);
								if (!modDate)
									modDate=creadate;
								if (modDate) {
									pNextName[0]='\t';
									CreateIsoDateString(modDate,pNextName+1);
									pNextName+=StrLen(pNextName);
								}
							}
							StrCopy(pNextName,"\r\n");
							pNextName+=2;
						}
					}
					currdatabase++;
					if (currdatabase>=numdatabases)
					dirIterator=vfsIteratorStop;
				}
				blocksize=pNextName-block;	// no #0 at end!
				MemPtrFree(FileName);
			} else
				break;
		} else if (isadir) {
			Char *FileName2=MemPtrNew(260);
			block[0]=0;
			blocksize=0;
			if (FileName2) {
				char* pEndBlock;
				char* pNextName=block;
				block[0]=0;
				blocksize=1024;	//maxblocksize;
				pEndBlock=block+blocksize-262;	// if more files, continue in next block
				while (dirIterator!=vfsIteratorStop && pNextName<pEndBlock) {
					FileInfoType info;
					info.nameP=FileName2;//point to local buffer
					info.nameBufLen=260;
					err=VFSDirEntryEnumerate(FileHandle,&dirIterator,&info);
					if (err==errNone){
						FileRef FileHandle2;
						UInt32 TotalBytes,modDate;
						char fullname[260];

						StrCopy(pNextName,FileName2);
						pNextName+=StrLen(FileName2);
						if (info.attributes & vfsFileAttrDirectory) {
							StrCopy(pNextName,"\\");
							pNextName++;
						}
						StrCopy(fullname,FileName+3);
						if (fullname[0]==0 || fullname[StrLen(fullname)-1]!='/')
							StrNCat(fullname,"/",sizeof(fullname)-1);
						StrNCat(fullname,FileName2,sizeof(fullname)-1);
						err=VFSFileOpen(VolRefNumCurrent, fullname, vfsModeRead, &FileHandle2);
						if (err==errNone) {
							if (info.attributes & vfsFileAttrDirectory) {
								err=errNone;
								TotalBytes=0;
							} else
								err=VFSFileSize(FileHandle2,&TotalBytes);
							if (err==errNone) {
								pNextName[0]='\t';
								StrIToA(pNextName+1,TotalBytes);
								pNextName+=StrLen(pNextName);
								modDate=0;
								err=VFSFileGetDate(FileHandle2, vfsFileDateModified, &modDate);
								if (err) // no mod date - use creation date!
									err=VFSFileGetDate(FileHandle2, vfsFileDateCreated, &modDate);
								pNextName[0]='\t';
								if (err==errNone && modDate) {
									CreateIsoDateString(modDate,pNextName+1);
									pNextName+=StrLen(pNextName);
								} else {
									pNextName++;
								}
								pNextName[0]='\t';
								StrIToA(pNextName+1,info.attributes);
								pNextName+=StrLen(pNextName);
							}
							VFSFileClose(FileHandle2);
						}
						StrCopy(pNextName,"\r\n");
						pNextName+=2;
					} else {
						dirIterator=vfsIteratorStop;
					}
				}
				blocksize=pNextName-block; // without #0 at end!
				MemPtrFree(FileName2);
			} else
				break;
		} else {
			blocksize=FileSize-offset;
			if (blocksize>maxblocksize)
				blocksize=maxblocksize;
			if (blocksize>0) {
				err=VFSFileRead(FileHandle,blocksize,block,&numBytesRead);
				if (err)
					return err_file;
			}
		}
//	divide the read data into smaller blocks
		sendptr=block;
		while (blocksize>0) {
			sendblocksize=blocksize;
			if (!isUSB && sendblocksize>1024)
				sendblocksize=1024;

			err=SendNextBlockRetry10(CMD_BODY,(char*)sendptr,sendblocksize,sequencenr);

			if (sequencenr==255)
				sequencenr=0;
			else
				sequencenr++;
			offset += sendblocksize;
			blocksize-=sendblocksize;
			sendptr+=sendblocksize;
			if (err!=RQ_OK) {
				AddStringToHistory("Body block not acknowledged!");
				  if (err==RQ_LOCALABORT)
					return err_abort;
				else
					return err_recv;
			}
			if (FileSize!=0xFFFFFFFF)
				ShowPercentDone(offset, FileSize);
			else {
				StrCat(progress,"#");
				if (StrLen(progress)>10)
					progress[0]=0;
				ShowStatus(progress);
			}
		}
	}

	// Send end of file block
	err=SendNextBlockRetry10(CMD_EOF,NULL,0,0);
	if (err!=RQ_OK) {	// timeout!
		AddStringToHistory("End of file block not acknowledged!");
		  if (err==RQ_LOCALABORT)
			return err_abort;
		else
			return err_recv;
	}
	return 0;
}

// ******************* AddCommandToHistory ********************
// *** Adds a command with parameter to the history listbox ***
// ************************************************************

void AddCommandToHistory(char* command,char* textstring)
{
	char buffer[100];
	StrLCopy(buffer,command,sizeof(buffer)-1);
	StrNCat(buffer," ",sizeof(buffer)-1);
	StrNCat(buffer,textstring,sizeof(buffer)-1);
	AddStringToHistory(buffer);
}

// ******************** AddStringToHistory ********************
// ******** Adds a text string to the history listbox *********
// ************************************************************

void AddStringToHistory(char* textstring)
{
	char *olditem;
	char *historyitem;
	ListPtr lst;
	int i;

	lst=FrmGetObjectPtr(mainform,FrmGetObjectIndex(mainform, ID_HISTORY));
	if (!lst)
		return;

	historyitem=(char*)MemPtrNew(StrLen(textstring)+1);
	if (!historyitem)
		return;
	StrCopy(historyitem,textstring);
	if (historycount==historylen) {
		olditem=historylist[historylen-1];
		if (olditem)
			MemPtrFree(olditem);
	}
	if (historycount<historylen)
		historycount++;

	for (i=historycount-2;i>=0;i--)
		historylist[i+1]=historylist[i];
	historylist[0]=historyitem;
	
	LstSetListChoices(lst,historylist,historycount);
	LstDrawList(lst);
}

// ********************* ShowPercentDone **********************
// ** Shows how many percent of the transfer have been done ***
// ************************************************************

static int lastpercent=-1;

void ShowPercentDone(Int32 current,Int32 all)
{
	char buffer[50],numbuf[10];
	int percent;

	if (all==0)
		return;
	percent=(100*current)/all;
	if (percent<0) percent=0;
	if (percent>100) percent=100;
	if (lastpercent==percent)
		return;
	StrCopy(buffer,"Status: ");
	StrIToA(numbuf,percent);
	StrNCat(buffer,numbuf,sizeof(buffer)-1);
	StrNCat(buffer," %",sizeof(buffer)-1);
	SafeSetLabel(ID_STATUS,buffer);
	lastpercent=percent;
}

// ********************* SetButtonText ************************
// *********** Sets the text of the left button ***************
// ************************************************************

static char newbuttonlabelbuffer[32];	// must remain available!!!
static char newportbuttonlabelbuffer[32];

void SetButtonText(char* buttontext)
{
	ControlPtr control;
	StrLCopy(newbuttonlabelbuffer,buttontext,sizeof(newbuttonlabelbuffer)-1);
	control=FrmGetObjectPtr(mainform,FrmGetObjectIndex(mainform,ID_STARTSTOP));
	if (control)
		CtlSetLabel(control,newbuttonlabelbuffer);
}

// ******************* SetPortButtonText **********************
// *********** Sets the text of the right button **************
// ************************************************************

void SetPortButtonText(char* buttontext)
{
	ControlPtr control;
	StrLCopy(newportbuttonlabelbuffer,buttontext,sizeof(newportbuttonlabelbuffer)-1);
	control=FrmGetObjectPtr(mainform,FrmGetObjectIndex(mainform,ID_CONFIGURE));
	if (control)
		CtlSetLabel(control,newportbuttonlabelbuffer);
}

// *********************** ShowStatus *************************
// *********** Sets the text of the status line ***************
// ************************************************************

void ShowStatus(char* statustext)
{
	char buffer[100];
	StrCopy(buffer,"Status: ");
	StrNCat(buffer,statustext,sizeof(buffer)-1);
	SafeSetLabel(ID_STATUS,buffer);
	lastpercent=-1;
}

// *********************** ShowName ***************************
// ************ Sets the text of the name line ****************
// ************************************************************

void ShowName(char* nametext)
{
	char buffer[100];
	StrCopy(buffer,"Name: ");
	StrNCat(buffer,nametext,sizeof(buffer)-1);
	SafeSetLabel(ID_NAME,buffer);
}

// ********************** SafeSetLabel ************************
// * Replaces the text of a label control. Warning: The new   *
// * must under no circumstances be longer than the previous  *
// * text, otherwise the program will crash!                  *
// ************************************************************

void SafeSetLabel(UInt16 id,char* labeltext)
{
	char buffer[100];
	char *p;
	int maxlen;

	MemSet(buffer, sizeof(buffer), ' ');	// fill with spaces!

	p=(char*)FrmGetLabel(mainform,id);
	if (!p)
		return;
	maxlen=StrLen(p);	// Set new string with exactly the same length!
	if (maxlen>99)
		maxlen=99;
	StrLCopy(buffer,labeltext,maxlen);
	buffer[StrLen(buffer)]=' ';	// don't cut off after name
	buffer[maxlen]=0;
	FrmCopyLabel(mainform, id, buffer);
}

// ********************* SerialOpenPort ***********************
// ************************************************************

typedef enum { 
	btLibUuidSize16 = 2, 
	btLibUuidSize32 = 4, 
	btLibUuidSize128 = 16 
} BtLibSdpUuidSizeEnum;

typedef struct BtLibSdpUuidType { 
	BtLibSdpUuidSizeEnum size; 
	UInt8 UUID[16]; 
} BtLibSdpUuidType; 

typedef struct { 
    BtLibSdpUuidType uuid; 
    const Char *name; 
} BtVdOpenParamsServer;

#define btLibDeviceAddressSize 6 
 
typedef struct BtLibDeviceAddressType { 
    UInt8 address[btLibDeviceAddressSize]; 
} BtLibDeviceAddressType; 

typedef UInt8 BtLibRfCommServerIdType;

typedef enum {
    btVdClient=0,        // initiates baseband and RFCOMM connections
    btVdServer           // waits for incoming baseband and RFCOMM connections
} BtVdRole;

typedef enum {
    btVdUseUuidList=0,   // use SDP to look for any one of the given UUIDs
    btVdUseChannelId     // just use the given RFCOMM channel id
} BtVdClientMethod;

typedef struct {
    UInt8                       len;           // length of table == number of UUIDs
    BtLibSdpUuidType*           tab;           // table of UUIDs
} BtVdUuidList;

typedef struct { 
    BtLibDeviceAddressType remoteDevAddr; 
    BtVdClientMethod method; 
    union { 
        BtLibRfCommServerIdType channelId; 
        BtVdUuidList uuidList; 
    } u; 
} BtVdOpenParamsClient;

typedef struct { 
    BtVdRole role; 
    union { 
        BtVdOpenParamsClient client; 
        BtVdOpenParamsServer server; 
    } u; 
    Boolean authenticate; 
    Boolean encrypt; 
} BtVdOpenParams; 

Err SerialOpenPort(UInt16 *SerialPortHandle,Int32 SerialSpeed)
{
	UInt32 value;
	Err err;
	BOOL NewSerialManagerV2Present;

	// From "Palm OS Companion2.pdf"
	// To check whether you can use the new Serial Manager API, check
	// for the existence of the new Serial Manager feature set by calling
	// FtrGet as follows:
	err = FtrGet(sysFileCSerialMgr,sysFtrNewSerialPresent,&value);

	// If the new Serial Manager is installed, the value parameter is non-zero
	// and the returned error is zero(for no error).*/
	NewSerialManagerPresent=(err==0 && value!=0);

	if (NewSerialManagerPresent) {
		// To check for the existence of version 2 of the new Serial Manager,
		// you should check both the Serial Manager version number and the
		// Palm OS version number as follows:
		err = FtrGet(sysFileCSerialMgr,sysFtrNewSerialVersion,&value);

		// If the value parameter is 2, the romVersion is 0x04003000, and
		// both calls to FtrGet return 0(for no error), version 2 of the new
		// Serial Manager feature set is present.
		NewSerialManagerV2Present=(err==0 && value>=2 && romversion>=0x04000000);
	} else {
		NewSerialManagerV2Present=false;
		UseVersion1SendRecv=romversion<0x02000000;
	}

	// Now, try to open the port

	if (isHandspring && isUSB) {
		if (NewSerialManagerPresent) {
			AddStringToHistory("Using Handspring USB call");
			err = SrmOpen( usbLibSrmDriverIDDesktop, SerialSpeed, SerialPortHandle );
			if (err)
				return err;
		} else {
			AddStringToHistory("Using old Handspring USB library");
			err=SysLibFind("USB Library",SerialPortHandle);
			if (err) {
				AddStringToHistory("USB library not found!");
				return err;
			}
			err=SerOpen(*SerialPortHandle,0,SerialSpeed);
			if (err==serErrAlreadyOpen) {
				AddStringToHistory("USB SerOpen failed!");
				SerClose(*SerialPortHandle);
			}
			if (err)
				return err;
		}
	} else if (NewSerialManagerV2Present) {
		SrmOpenConfigType conf;
		BtVdOpenParams btParams;
		UInt32 portid;
		AddStringToHistory("Using new serial manager v2");
		MemSet(&conf, sizeof(conf), 0);			// fill with spaces!
		conf.baud=SerialSpeed;
		portid=serPortLocalHotSync;
		if (isBlueTooth) {
			MemSet( &btParams, sizeof(btParams), 0 );     // note this leaves the remote dev addr null
			btParams.role = btVdClient;                   // we are the client side
			btParams.authenticate = true; //link authentication
			btParams.encrypt = true;      //link encryption

			btParams.u.client.method = btVdUseChannelId;
			btParams.u.client.u.channelId = 0x53;

			// Set up the SrmExtOpen param block to reference our bluetooth param block
			MemSet( &conf, sizeof(conf), 0 );
			conf.function = 0;                       // must be zero
			conf.drvrDataP = (MemPtr)&btParams;      // bluetooth-specific params go here
			conf.drvrDataSize = sizeof(btParams);    // size of bluetooth-specific params
			portid=sysFileCVirtRfComm;
		} else
			conf.function=ThisPrograms4byteCode;	// needed for USB!
		err=SrmExtOpen(portid,&conf,sizeof(conf),SerialPortHandle);
		if (err)
			return err;
	} else if (NewSerialManagerPresent) {
		AddStringToHistory("Using new serial manager v1");
		err = SrmOpen(serPortLocalHotSync,SerialSpeed,SerialPortHandle);
		if (err)
			return err;
	} else {	// old serial manager
		// From "Palm OS Companion2.pdf"
		AddStringToHistory("Using old serial manager");
		err=SysLibFind("Serial Library",SerialPortHandle);
		if (err)
			return err;
		err=SerOpen(*SerialPortHandle,0,SerialSpeed);
		if (err==serErrAlreadyOpen)
			SerClose(*SerialPortHandle);
		if (err)
			return err;
	}

	// Now the port is open - configure it!
	// For baud rates above 19200, the use of full hardware
	// handshaking(srmSettingsFlagRTSAutoM |
	// SrmSettingsFlagCTSAutoM) is advised.
	err=0;
	
	if (NewSerialManagerPresent) {
		UInt16 paramSize;
		UInt32 flags = srmSettingsFlagBitsPerChar8 | srmSettingsFlagStopBits1 |
				srmSettingsFlagRTSAutoM | srmSettingsFlagCTSAutoM |
				srmSettingsFlagFlowControlIn;
		Int32 ctsTimeout = 5*PalmSysTicksPerSecond(); // partial sends not working, so wait long enough!
		paramSize = sizeof(flags);
		err = SrmControl(*SerialPortHandle,srmCtlSetFlags,&flags,&paramSize);
		if (!err) {
			paramSize = sizeof(ctsTimeout);
			err = SrmControl(*SerialPortHandle,srmCtlSetCtsTimeout,&ctsTimeout,&paramSize);
		}
		if (err) {
			SrmClose(*SerialPortHandle);
			*SerialPortHandle=0;
		}
	} else {
		SerSettingsType serSettings;
		serSettings.baudRate = SerialSpeed;
		serSettings.flags = srmSettingsFlagBitsPerChar8 |
				srmSettingsFlagStopBits1 | srmSettingsFlagRTSAutoM |
				srmSettingsFlagCTSAutoM | srmSettingsFlagFlowControlIn;
		serSettings.ctsTimeout = 5*PalmSysTicksPerSecond();	// partial sends not working, so wait long enough!
		err = SerSetSettings(*SerialPortHandle, &serSettings);
		if (err) {
			SerClose(*SerialPortHandle);
			*SerialPortHandle=0;
		}
	}
	return err;
}

// *********************** SerialSend *************************
// ************************************************************

UInt32 SerialSend(UInt16 portId,const void *bufP,UInt32 count,Err *errP)
{
	if (NewSerialManagerPresent) {
		return SrmSend(portId,bufP,count,errP);
	} else if (UseVersion1SendRecv) {
		Err err=SerSend10(portId,bufP,count);
		*errP=err;
		if (err==0)
			return count;
		else
			return 0;
	} else {
		return SerSend(portId,bufP,count,errP);
	}
}

// ********************** SerialReceive ***********************
// ************************************************************

UInt32 SerialReceive(UInt16 portId,void *rcvBufP,UInt32 count,Int32 timeout,Err *errP)
{
	if (NewSerialManagerPresent) {
		return SrmReceive(portId,rcvBufP,count,timeout,errP);
	} else if (UseVersion1SendRecv) {
		Err err=SerReceive10(portId,rcvBufP,count,timeout);
		*errP=err;
		if (err==0)
			return count;
		else
			return 0;
	} else {
		return SerReceive(portId,rcvBufP,count,timeout,errP);
	}
}

// ******************** SerialReceiveWait *********************
// ************************************************************

Err SerialReceiveWait(UInt16 portId,UInt32 bytes,Int32 timeout)
{
	if (NewSerialManagerPresent) {
		return SrmReceiveWait(portId,bytes,timeout);
	} else {
		return SerReceiveWait(portId,bytes,timeout);
	}
}

// ********************** SerialClearErr **********************
// ************************************************************

Err SerialClearErr(UInt16 portId)
{
	if (NewSerialManagerPresent) {
		return SrmClearErr(portId);
	} else {
		return SerClearErr(portId);
	}
}

// *********************** SerialClose ************************
// ************************************************************

Err SerialClose(UInt16 portId)
{
	if (NewSerialManagerPresent) {
		return SrmClose(portId);
	} else {
		return SerClose(portId);
	}
}

// ****************** SerialSetReceiveBuffer ******************
// ************************************************************

Err SerialSetReceiveBuffer(UInt16 portId,void *bufP,UInt16 bufSize)
{
	if (NewSerialManagerPresent) {
		return SrmSetReceiveBuffer(portId,bufP,bufSize);
	} else {
		return SerSetReceiveBuffer(portId,bufP,bufSize);
	}
}

// ********************* SerialSendFlush **********************
// ************************************************************

Err SerialSendFlush(UInt16 portId)	// delete any unsent data
{
	if (NewSerialManagerPresent) {
		return SrmSendFlush(portId);
	} else {
		return SerSendFlush(portId);
	}
}

Err SerialClearReceiveBuffer()
{
	Err err;
	char flushbuf[64];
	UInt32 receive;
	do {	//Flush
		receive=sizeof(flushbuf);
		err=TimedReceiveBytes(flushbuf,&receive,100);
		if (err==err_abort)
			return CMD_ABORT;
	} while (err==0);
	do {	//Flush
		err=TimedReceiveChar(flushbuf,100);
		if (err==err_abort)
			return CMD_ABORT;
	} while (err==0);
	return err;
}
