// Function prototypes

static UInt32 PalmSysTicksPerSecond();

// Window handling functions
UInt32 PilotMain (UInt16 cmd,void *cmdPBP,UInt16 launchFlags);
static UInt16 StartApplication (void);
static void StopApplication (void);
static void EventLoop (void);
BOOL ApplicationHandleEvent (EventType *event);
Boolean MainFormEventHandler (EventType *eventP);
void SetPortButtonAndHistory();
BOOL ShowConfigurePortForm();

// File Upload+Download
Err SerialReceiveFile(FileRef FileHandle,char* FileName,char* block,UInt32 restartat,UInt32* seekto);
Err SerialSendFile(FileRef FileHandle,char* FileName,char* block, UInt32 restartat, BOOL sizeonly, BOOL isadir);

// Block send/receive calls
char ReceiveNextBlock(char* recvbody,UInt32 maxbodysize,UInt32* datareceived,UInt8* sequencenr);
int SendNextBlock(char request,char* databuf,UInt32 datalen,UInt8 sequencenr);

// Serial encapsulation
Err TimedReceiveChar (char* ch,Int32 timeout_msec);
Err TimedReceiveBytes(char* recvbuf,UInt32* numbytes,Int32 timeout_msec);
Err SendChar(char ch);
Err SendBytes(char* sendbuf,Int32 numbytes);
void MainReceiveFunction(void);

// UI calls
void ShowPercentDone(Int32 current,Int32 all);
void SetButtonText(char* buttontext);
void SetPortButtonText(char* buttontext);
void ShowStatus(char* statustext);
void ShowName(char* nametext);
void SafeSetLabel(UInt16 id,char* labeltext);
void AddStringToHistory(char* textstring);
void ListDrawFunction(Int16 itemNum,RectangleType *bounds,Char **itemsText);
void AddCommandToHistory(char* command,char* textstring);

// Serial manager calls
Err SerialOpenPort(UInt16 *SerialPortHandle,Int32 SerialSpeed);
UInt32 SerialSend(UInt16 portId,const void *bufP,UInt32 count,Err *errP);
UInt32 SerialReceive(UInt16 portId,void *rcvBufP,UInt32 count,Int32 timeout,Err *errP);
Err SerialReceiveWait(UInt16 portId,UInt32 bytes,Int32 timeout);
Err SerialClearErr(UInt16 portId);
Err SerialClose(UInt16 portId);
Err SerialSetReceiveBuffer(UInt16 portId,void *bufP,UInt16 bufSize);
Err SerialSendFlush(UInt16 portId);
Err SerialClearReceiveBuffer();

// Constants: Use bit patterns which differ between each other by at least 2 bits!

#define RQ_OK  27          // binary 0011011      // block OK
#define RQ_FAILED 73       // binary 1001001      // block OK, operation failed
#define RQ_TRANSFERERR 85  // binary 1010101      // block transfer error
#define RQ_READY 119       // binary 1110111      // signal "transfer ready"
#define RQ_CRCERR 99       // binary 1100011      // block has bad CRC
#define RQ_WAIT 60         // binary 0111100      // tell the client to wait an extended time - for operations taking more than 5 seconds
// these two are never sent across the line!
#define RQ_LOCALABORT 0    // binary 0000000      // user aborted operation - never sent across line! Use RQ_FAILED instead
#define RQ_LOCALTIMEOUT 1  // binary 0000001      // Send or receive call timed out -> retry after 5 seconds delay

// commands are uppercase, data blocks lowercase, errors are numbers

#define CMD_LIST    'L'
#define CMD_GET     'G'
#define CMD_SEND    'S'
#define CMD_DELETE  'D'
#define CMD_RENFR   'F'
#define CMD_RENTO   'T'
#define CMD_COPYTO  'O'
#define CMD_MKDIR   'M'
#define CMD_CONNECT 'C'
#define CMD_SETATTR 'A'
#define CMD_SETTIME 'I'
#define CMD_RESTART 'R'
#define CMD_GETSIZE 'Q'
#define CMD_OVERWRITE 'W'

#define CMD_SIZE    's'
#define CMD_BODY    'b'
#define CMD_EOF     'e'

#define CMD_ABORT   '1'
#define CMD_RECVERR '2'
#define CMD_CRCERR  '3'
#define CMD_TIMEOUT '4'

#define err_abort 1
#define err_recv 2
#define err_send 3
#define err_mem 4
#define err_file 5

#define defmaxblockpart 8192
#define defmaxblocksize 64512     //63*1024 = 64k-1k

typedef struct _DBWR {
	char* block;
	UInt32 blockbytecontent;
	UInt8 sequencenr;
	char progress[12];
	UInt32 FileSize,offset,restartdiscard;
	BOOL eofreached;
} DBWR,*pDBWR;

