unit Filer;

{
 ***** version 1.01 ***** 
 (c) Pavel Zampach (zampach@volny.cz), 2003
 Source code for Borland Delphi Compiler (originally ver. 7.0)
}

interface

const
  PACKET_DATA_SIZE = $0800;
  FNAME_LENGTH = 13;

//   Filer communications commands
  CMD_ASK_DIR           = $12;
  CMD_CONNECT_SERVER    = $40;
  CMD_DATA_END          = $01;
  CMD_DEL_DIR           = $0E;
  CMD_DEL_FILE          = $04;
  CMD_DISCONNECT_SERVER = $41;
  CMD_GET_DATA          = $0C;
  CMD_GET_DIR           = $09;
  CMD_GET_FILENAME      = $0B;
  CMD_INIT_GET          = $16;
  CMD_MAKE_DIR          = $0A;
  CMD_REN_FILE          = $0D;
  CMD_SEND_DATA         = $15;
  CMD_SEND_FILENAME     = $03;
  CMD_SEND_PATH         = $08;
 
type
  TData = array [1..PACKET_DATA_SIZE] of byte;
  TPData = ^TData;

  TDirEntry = packed record
    fattr  : byte;
    ftime  : word;
    fdate  : word;
    fsize  : longword;
    fname  : array [0..FNAME_LENGTH-1] of char;
  end;

var
//  Data for packet traffic
  Packet :  record
    size     : word;
    size2    : word;
    data     : TPData;
    data2    : TPData;
    endOfDir : boolean;
  end;
       
  Connected : boolean;

function FilerReset : boolean;
function FilerRequest (Funct : byte) : boolean; overload;
function FilerRequest (Funct : byte; Data : Pchar) : boolean; overload;
function FilerConnect (const CommPort : string; const CommSpeed : integer) : boolean;
function FilerDisconnect : boolean;

{ ****************************************************************** }

implementation

uses
  SysUtils,
  Windows;

const
  NUL         = 0;
  SOH         = 1;
  STX         = 2;
  STATUS_OK   = $81;
  MAX_PKT_LEN = PACKET_DATA_SIZE*2+100;
  SIG_LEN     = 5;
  
// packet signature
  Signature : array [1..SIG_LEN] of byte = ($16, $16, $16, $10, $02);

type
  TWorkBuf = array [1..MAX_PKT_LEN] of byte;

var
  hCommFile : THandle;
  PacketFunct, PacketCount, PacketStatus : byte;

{ ------------------------------------------------------------------ }

function UpdateCRC16 (CRC : word; Data : byte) : word;
var
  i : byte;
begin
  Result := CRC xor Data;
  for i := 1 to 8 do
    if (Result and 1) = 1 then
      Result := (Result shr 1) xor $A001
    else
      Result := Result shr 1;
end;  // function UpdateCRC16

{ ------------------------------------------------------------------ }

procedure SendPacket;

var
  Checksum, i, WorkSize : word;
  WorkBuf : TWorkBuf;
  BytesWritten : LongWord;

{ -------------- }

procedure SendByte (value : byte; Chsum : boolean);

begin
  inc (WorkSize);
  WorkBuf[WorkSize] := value;
// '$10' CRC trick, got to send it twice
  if (value = $10) and Chsum then begin
    inc (WorkSize);
    WorkBuf[WorkSize] := value;
  end;
// update checksum
  if Chsum then
    Checksum := UpdateCRC16 (Checksum, value);
end;  // function SendByte

{ -------------- }

begin  // function SendPacket
  Checksum := 0;
  WorkSize := 0;
  
// synchonization - 4 NUL bytes
  if not Connected then
    for i := 1 to 4 do
      SendByte (NUL, false);
      
// send signature, no CRC yet
  for i := 1 to SIG_LEN do
    SendByte (Signature[i], false);
  SendByte (SOH, true);
  SendByte (PacketFunct, true);
  SendByte (PacketCount, true);
// send Data marker
  SendByte (SOH, true);
  SendByte (STX, true);

  if PacketFunct = CMD_SEND_DATA then
  begin
    SendByte (0, true);
    SendByte (0, true);
    SendByte (Lo (Packet.size), true);
    SendByte (Hi (Packet.size), true);
    for i := 1 to Packet.size do
      SendByte (Packet.data^[i], true);
  end;

  if (PacketFunct = CMD_MAKE_DIR) or
     (PacketFunct = CMD_DEL_DIR) or
     (PacketFunct = CMD_DEL_FILE) or
     (PacketFunct = CMD_SEND_PATH) or
     (PacketFunct = CMD_SEND_FILENAME) or
     (PacketFunct = CMD_GET_FILENAME) or
     (PacketFunct = CMD_REN_FILE) or
     (PacketFunct = CMD_ASK_DIR) then
  begin
// send path/filename (FROM) size LO, HI
    SendByte (Lo (Packet.size), true);
    SendByte (Hi (Packet.size), true);
// send path/filename (FROM)
    for i := 1 to Packet.size do
      SendByte (Packet.data^[i], true);
// send Data End Marker
    SendByte (NUL, true);
    SendByte (NUL, true);
  end;
  
  if (PacketFunct = CMD_REN_FILE) then
  begin
// send filename TO size LO, HI
    SendByte (Lo (Packet.size2), true);
    SendByte (Hi (Packet.size2), true);
// send filename TO
    for i := 1 to Packet.size2 do
      SendByte (Packet.data2^[i], true);
// send Data End Marker
    SendByte (NUL, true);
    SendByte (NUL, true);
  end;

  if PacketFunct = CMD_GET_DATA then
  begin
// double word size, MSB always zero
    SendByte (0, true);
    SendByte (0, true);
    SendByte (Lo (Packet.size), true);
    SendByte (Hi (Packet.size), true);
  end;

  if PacketFunct = CMD_DATA_END then
    SendByte (NUL, true);

// ----- COMMON PART -----
// No CRC on this one (tricky!)
  SendByte ($10, false);
  SendByte ($03, true);
// Finally send Checksum LO, HI
  SendByte (Lo (Checksum), false);
  SendByte (Hi (Checksum), false);

// really send packet from buffer
  WriteFile (hCommFile, WorkBuf, WorkSize, BytesWritten, nil);
end;  // function SendPacket

{ ------------------------------------------------------------------ }

function GetPacket : boolean;

var
  c, Data, ByteHi, ByteLo : byte;
  i, Checksum, EstimatePktLen, WorkPtr  : word;
  BytesRead : longword;
  WorkBuf : TWorkBuf;

{ -------------- }

function GetByte (var value : byte; Chsum : boolean) : boolean;

begin
  Result := false;
  
  if WorkPtr = BytesRead then exit;
  inc (WorkPtr);
  value := WorkBuf[WorkPtr];
// '$10' CRC trick, receive it twice
  if (value = $10) and Chsum then begin
    if WorkPtr = BytesRead then exit;
    inc (WorkPtr);
    value := WorkBuf[WorkPtr];
  end;
// update checksum
  if Chsum then
    Checksum := UpdateCRC16 (Checksum, value);

  Result := true;
end;  // function GetByte

{ -------------- }

begin
  Checksum := 0;
  Result := false;
  WorkPtr := 0;
  if PacketFunct = CMD_GET_DATA then
    EstimatePktLen := MAX_PKT_LEN
  else
    EstimatePktLen := 100;

// really read packet into buffer, end by timeout
  ReadFile (hCommFile, WorkBuf, EstimatePktLen, BytesRead, nil);

// search signature
  c := 0;
  repeat
    if not GetByte (Data, false) then exit;
    inc (c);
    if Data <> Signature[c] then c := 0;
  until c = SIG_LEN;

// get SOH
  if not GetByte (Data, true) then exit;
  if Data <> SOH then exit;
  if not GetByte (PacketFunct, true) then exit;
  if not GetByte (PacketCount, true) then exit;
  if not GetByte (PacketStatus, true) then exit;
// get Data marker
  if not GetByte (Data, true) then exit;
  if Data <> STX then exit;

  if (PacketFunct = CMD_CONNECT_SERVER) or
     (PacketFunct = CMD_DISCONNECT_SERVER) then
  begin
// Data end marker
    if not GetByte (Data, true) then exit;
    if Data <> NUL then exit;
    if not GetByte (Data, true) then exit;
    if Data <> NUL then exit;
  end;

  if (PacketFunct = CMD_MAKE_DIR) or
     (PacketFunct = CMD_DEL_DIR) or
     (PacketFunct = CMD_DEL_FILE) or
     (PacketFunct = CMD_REN_FILE) or
     (PacketFunct = CMD_SEND_PATH) or
     (PacketFunct = CMD_SEND_FILENAME) or
     (PacketFunct = CMD_GET_FILENAME) or
     (PacketFunct = CMD_SEND_DATA) then
  begin
// get Data marker and size MSB (always 0)
    for i := 1 to 4 do begin
      if not GetByte (Data, true) then exit;
      if Data <> NUL then exit;
    end;
// get Data size
    if not GetByte (ByteLo, true) then exit;
    if not GetByte (ByteHi, true) then exit;
    Packet.size := (ByteHi * $100) + ByteLo;
  end;

  if (PacketFunct = CMD_DATA_END) or
     (PacketFunct = CMD_INIT_GET) or
     (PacketFunct = CMD_ASK_DIR) then
// get Data size & end marker (6 bytes)
     for i := 1 to 6 do begin
       if not GetByte (Data, true) then exit;
       if Data <> NUL then exit;
     end;

  if PacketFunct = CMD_GET_DATA then
  begin
// get Data marker and size MSB (always 0)
     for i := 1 to 4 do begin
       if not GetByte (Data, true) then exit;
       if Data <> NUL then exit;
     end;
// get Data size
     if not GetByte (ByteLo, true) then exit;
     if not GetByte (ByteHi, true) then exit;
     Packet.size := (ByteHi * $100) + ByteLo;
     for i := 1 to Packet.size do
       if not GetByte (Packet.data^[i], true) then exit;
  end;

  if PacketFunct = CMD_GET_DIR then
  begin
// get Data marker and size MSB (always 0)
    for i := 1 to 4 do begin
      if not GetByte (Data, true) then exit;
      if Data <> NUL then exit;
    end;
// get Data size
    if not GetByte (ByteLo, true) then exit;
    if not GetByte (ByteHi, true) then exit;
    Packet.size := (ByteHi * $100) + ByteLo;
// get Data
    for i := 1 to Packet.size do
      if not GetByte (Packet.data^[i], true) then exit;
// special 'dir' Data end marker
    if not GetByte (Data, true) then exit;
    if Data > 1 then exit;
    Packet.endOfDir := (Data = NUL);

    for i := 1 to 3 do begin
      if not GetByte (Data, true) then exit;
      if Data <> NUL then exit;
    end;
  end;

// ----- COMMON PART -----
// get CRC Marker (no CRC on the next byte!)
  if not GetByte (Data, false) then exit;
  if Data <> $10 then exit;
  if not GetByte (Data, true) then exit;
  if Data <> $03 then exit;
// get received CRC
  if not GetByte (ByteLo, false) then exit;
  if not GetByte (ByteHi, false) then exit;
// check if CRC is good
  if ((ByteHi * $100) + ByteLo) <> Checksum then exit;
  Result := true;
end; // function GetPacket

{ ------------------------------------------------------------------ }

function FilerRequest (Funct : byte) : boolean;  overload;

const
  MAX_ATT = 3;

var
  Att, Count : byte;
  BGetPacket : boolean;

begin
  Result := false;
  if Packet.size > PACKET_DATA_SIZE then exit;
  if (Funct = CMD_CONNECT_SERVER) or
     (Funct = CMD_DISCONNECT_SERVER) then PacketCount := 0;
  Count := PacketCount;
  PacketFunct := Funct;

// packet handshake
  Att := 0;
  repeat
    PurgeComm (hCommFile, PURGE_TXCLEAR or PURGE_RXCLEAR);
    SendPacket;
    inc (Att);
    BGetPacket := GetPacket;
  until BGetPacket or (Att = MAX_ATT);

// check received values
  if (not BGetPacket) or
     (PacketStatus <> STATUS_OK) or
     (PacketCount  <> Count) or
     (PacketFunct  <> Funct) then exit;

// cyclic update packet counter
  if PacketCount = $FF then
    PacketCount := 0
  else
    inc (PacketCount);
    
  Result := true;
end;  // funct FilerRequest

{ ------------------------------------------------------------------ }

function FilerRequest (Funct : byte; Data : Pchar) : boolean; overload;

begin
  Packet.data := TPData (Data);
  Packet.size := strLen (Data);
  Result := FilerRequest (Funct);
end;  // funct FilerRequest with filename

{ ------------------------------------------------------------------ }

function FilerReset : boolean;

begin
  sleep (300);
  PurgeComm (hCommFile, PURGE_TXABORT or PURGE_RXABORT or PURGE_TXCLEAR or PURGE_RXCLEAR);
  Result := FilerRequest (CMD_CONNECT_SERVER);
end;

{ ------------------------------------------------------------------ }

function FilerConnect (const CommPort : string; const CommSpeed : integer) : boolean;

const
  RX_BUF = $C00;
  TX_BUF = $C00;

  fBinary              = $00000001;
  fParity              = $00000002;
  fOutxCtsFlow         = $00000004;
  fOutxDsrFlow         = $00000008;
  fDtrControlEnable    = $00000010;
  fDtrControlHandshake = $00000020;
  fDsrSensitivity      = $00000040;
  fTXContinueOnXoff    = $00000080;
  fOutX                = $00000100;
  fInX                 = $00000200;
  fErrorChar           = $00000400;
  fNull                = $00000800;
  fRtsControlEnable    = $00001000;
  fRtsControlHandshake = $00002000;
  fAbortOnError        = $00004000;
  fDummy2              = $FFFF8000;

var
  dcbPort : TDCB;
  CommPortTimeouts : TCommTimeouts;
  SpeedCorr : word;

begin
  if Connected then begin
    Result := true;
    exit;
  end;

// open file for serial communications
  hCommFile := CreateFile (PChar(CommPort), GENERIC_READ or GENERIC_WRITE, 0, nil, OPEN_EXISTING, 0, 0);
  if hCommFile = INVALID_HANDLE_VALUE then begin
    Result := false;
    exit;
  end;
  PurgeComm (hCommFile, PURGE_TXABORT or PURGE_RXABORT or PURGE_TXCLEAR or PURGE_RXCLEAR);

// set port parameters
  if getCommState (hCommFile, dcbPort) then with dcbPort do begin
    BaudRate := CommSpeed;
    ByteSize := 8;
    Parity   := NOPARITY;
    StopBits := ONESTOPBIT;
    Flags    := 0;  // no handshake!
    setCommState (hCommFile, dcbPort);
  end;

// set port timeouts
  SpeedCorr := 10000 div CommSpeed;
  if getCommTimeouts (hCommFile, CommPortTimeouts) then with CommPortTimeouts do begin
    ReadIntervalTimeout         := 25 + SpeedCorr;
    ReadTotalTimeoutMultiplier  := 1 + SpeedCorr;
    ReadTotalTimeoutConstant    := 1000;
    WriteTotalTimeoutMultiplier := 1 + SpeedCorr;
    WriteTotalTimeoutConstant   := 1000;
    setCommTimeouts (hCommFile, CommPortTimeouts);
  end;

// set recieve and transmit buffer size
  SetupComm (hCommFile, RX_BUF, TX_BUF);

  Result := FilerRequest (CMD_CONNECT_SERVER);
  if not Result then begin
    sleep (300);
    PurgeComm (hCommFile, PURGE_TXABORT or PURGE_RXABORT or PURGE_TXCLEAR or PURGE_RXCLEAR);
    CloseHandle (hCommFile);
    exit;
  end;

  Connected := true;
end; // funct FilerConnect

{ ------------------------------------------------------------------ }

function FilerDisconnect :boolean;

begin
  if not Connected then begin
    Result := true;
    exit;
  end;

  Result := FilerRequest (CMD_DISCONNECT_SERVER);
  CloseHandle (hCommFile);
  Connected := false;
end;

{ ------------------------------------------------------------------ }

initialization
  Connected := false;
end.