library hp200lx;

{
 ***** version 1.02 ***** 
 (c) Pavel Zampach (zampach@volny.cz), 2003
 Source code for Borland Delphi Compiler (originally version 7.0)
}

{$E wfx}
{$IOChecks Off}
{$RangeChecks On}
{$Resource HP200LX.RES}

uses
  Windows,
  SysUtils,
  IniFiles, 
  ShellAPI,
  Fsplugin in 'FSPLUGIN.PAS',
  Filer in 'FILER.PAS';

const
  PLUGIN_TITLE = 'HPLX file system';
  HELLO_TITLE  = 'HPLX plugin 1.02, (c) P. Zampach, 2003';

type
  CharArr = array [0..MAX_PATH-1] of char;

var
  ProgressProc : tProgressProc;
  LogProc      : tLogProc;
  RequestProc  : tRequestProc;
  PluginNumber : integer;
  DiskNr       : byte;
  IsRoot       : boolean;
  SDrives      : string;
  IniFileName  : string;

{ ------------------------------------------------------------------ }

function trimFileName (const OldName : string) : string;

const
  DOSChar : set of char =
    ['A'..'Z', 'a'..'z', '0'..'9', '_', '-', '~', '`', '!', '@', '#', '$', '%', '^', '&', '(', ')', '{', '}', #39 ];

var
  i, q, PosPoint : integer;
  c : char;

begin
  Result := '';
  PosPoint := lastDelimiter ('.', OldName);
  if PosPoint = 0 then
    PosPoint := length (OldName) + 1;

  for i := 1 to PosPoint - 1 do begin
    c := OldName[i];
    if (c in DOSChar) and (length (Result) < 8) then
      Result := Result + c;
  end;

  if Result = '' then Result := '$$$';
  Result := Result + '.';
  q := length (Result) + 3;

  for i := (PosPoint + 1) to (length (OldName)) do begin
    c := OldName[i];
    if (c in DOSChar) and (length (Result) < q) then
      Result := Result + c;
  end;
end;

{ ------------------------------------------------------------------ }

procedure dirEntryToFindData (const DirEntry : TDirEntry; var FindData : tWIN32FINDDATA);

var
  FT : TFileTime;
  
begin
  fillChar (FindData, sizeOf (tWIN32FINDDATA), 0);

  FindData.dwFileAttributes := DirEntry.fattr;
  DosDateTimeToFileTime (DirEntry.fdate, DirEntry.ftime, FT);
  LocalFileTimeToFileTime (FT, FindData.ftLastWriteTime);
  FindData.nFileSizeLow := DirEntry.fsize;
  strPLCopy (FindData.cFileName, DirEntry.fname, FNAME_LENGTH);
end;

{ ------------------------------------------------------------------ }

procedure rootToFindData (var FindData : tWIN32FINDDATA);

begin
  fillchar (FindData, sizeOf (tWIN32FINDDATA), 0);

  FindData.dwFileAttributes := FILE_ATTRIBUTE_DIRECTORY;
  FindData.ftLastWriteTime.dwHighDateTime := $FFFFFFFF;
  FindData.ftLastWriteTime.dwLowDateTime  := $FFFFFFFE;
  strPLCopy (FindData.cFileName, SDrives[DiskNr] + ':', 2);
end;

{ ------------------------------------------------------------------ }

function getDirEntry (const RemoteName : CharArr; var dirEntry : TDirEntry) : boolean;

begin
  Result := false;
  
  if not FilerRequest (CMD_ASK_DIR, RemoteName) then begin
    FilerReset;
    exit;
  end;
  Packet.data := @dirEntry;
  if not FilerRequest (CMD_GET_DIR) then begin
    FilerReset;
    exit;
  end;
  
  Result := true;
end;

{ ****************************************************************** }

function FsInit (PluginNr : integer; pProgressProc : tProgressProc; pLogProc : tLogProc;
                pRequestProc : tRequestProc) : integer; stdcall;

begin
  ProgressProc := pProgressProc;
  LogProc      := pLogProc;
  RequestProc  := pRequestProc;
  PluginNumber := PluginNr;

  Result := 0;
end;

{ ------------------------------------------------------------------ }

function FsFindClose (Hdl : thandle) : integer; stdcall;

begin
// empty
  Result := 0;
end;

{ ------------------------------------------------------------------ }

function FsFindFirst (Path : pchar; var FindData : tWIN32FINDDATA) : thandle; stdcall;

var
  Buf       : CharArr;
  DirEntry  : TDirEntry;
  iniFile   : TIniFile;
  CommPort  : string;
  CommSpeed : integer;

begin


  if not Connected then begin  // Login
// read INI file
    iniFile   := TIniFile.Create (IniFileName);
    CommPort  := iniFile.ReadString  ('HPLX', 'Port', 'COM1');
    CommSpeed := iniFile.ReadInteger ('HPLX', 'Speed', 19200);
    SDrives   := iniFile.ReadString  ('HPLX', 'Drives', 'CD');
    iniFile.Free;
 
    LogProc (PluginNumber, MsgTYPE_CONNECT, 'CONNECT \');
    if not FilerConnect (CommPort, CommSpeed) then begin
      Result := INVALID_HANDLE_VALUE;
      SetLastError (ERROR_FILE_NOT_FOUND);
      LogProc (PluginNumber, MsgTYPE_IMPORTANTERROR, 'Connect failed');
      exit;
    end;
    LogProc (PluginNumber, MsgTYPE_OPERATIONCOMPLETE, 'HPLX connected');
    LogProc (PluginNumber, MsgTYPE_DETAILS, HELLO_TITLE);
  end;

  IsRoot := (Path = '\');
  Result := 0;
    
  if IsRoot then begin  // Root processing
    DiskNr := 1;
    rootToFindData (FindData);
  end
  else begin            // Normal disk directory
    strCopy (Buf, Path + 1);
    strCat  (Buf, '\*.*');
    if not getDirEntry (Buf, DirEntry) then begin
      Result := INVALID_HANDLE_VALUE;
      SetLastError (ERROR_FILE_NOT_FOUND);
      exit;
    end;
    if Packet.endOfDir then begin           // Empty dir!
      Result := INVALID_HANDLE_VALUE;
      SetLastError (ERROR_NO_MORE_FILES);    
    end;
    dirEntryToFindData (DirEntry, FindData);
  end;

end;   // function FsFindFirst

{ ------------------------------------------------------------------ }

function FsFindNext (Hdl : thandle; var FindData : tWIN32FINDDATA) : bool; stdcall;

var
  DirEntry : TDirEntry;

begin
  if IsRoot then begin
    if DiskNr = length (SDrives) then
      Result := false
    else begin
      inc (DiskNr);
      rootToFindData (FindData);
      Result := true;
    end;
  end
  else begin // no root
    Packet.data := @DirEntry;
    if not FilerRequest (CMD_GET_DIR) then begin
      Result := false;
      FilerReset;
      exit;
    end;
    dirEntryToFindData (DirEntry, FindData);
    Result := not Packet.endOfDir;
  end;
end;   // function FsFindNext

{ ------------------------------------------------------------------ }

function FsDisconnect (DisconnectRoot : pchar) : bool; stdcall;

begin
  Result := FilerDisconnect;
  if Result then
    LogProc (PluginNumber, MSGTYPE_DISCONNECT, 'HPLX disconnected');
end;

{ ------------------------------------------------------------------ }

function FsDeleteFile (RemoteName : pchar) : bool; stdcall;

begin
  inc (RemoteName);    // skip over leading backslash
  LogProc (PluginNumber, MsgTYPE_DETAILS, PChar ('Del ' + RemoteName));

  Result := FilerRequest (CMD_DEL_FILE, RemoteName);
  if not Result then FilerReset;
end;   // function FsDeleteFile

{ ------------------------------------------------------------------ }

function FsMkDir (RemoteDir : pchar) : bool; stdcall;

begin
  inc (RemoteDir);    // skip over leading backslash
  LogProc (PluginNumber, MsgTYPE_DETAILS, PChar ('MkDir ' + RemoteDir));

  Result := FilerRequest (CMD_MAKE_DIR, RemoteDir);
  if not Result then FilerReset;
end;   // function FsMkDir

{ ------------------------------------------------------------------ }

function FsRemoveDir (RemoteName : pchar) : bool; stdcall;

begin
  inc (RemoteName);    // skip over leading backslash
  LogProc (PluginNumber, MsgTYPE_DETAILS, PChar ('DelDir ' + RemoteName));

  Result := FilerRequest (CMD_DEL_DIR, RemoteName);
  if not Result then FilerReset;
end;   // function FsRemoveDir

{ ------------------------------------------------------------------ }

function FsGetFile (RemoteName, LocalName : pchar; CopyFlags : integer ;
                    RemoteInfo : pRemoteInfo) : integer; stdcall;

var
  FileLength, Transferred, Delivered : longword;
  PercentDone  : integer;
  WorkFile     : file;
  PBuffer      : TPData;

begin
  if (CopyFlags and FS_COPYFLAGS_OVERWRITE) = 0 then
    if fileExists (LocalName) then begin
      Result := FS_FILE_EXISTS;
      exit
    end;  

  inc (RemoteName);    // skip over leading backslash
  LogProc (PluginNumber, MsgTYPE_DETAILS, PChar ('Copy ' + RemoteName + ' to ' + LocalName));

  if not FilerRequest (CMD_GET_FILENAME, RemoteName) then begin
    Result := FS_FILE_NOTFOUND;
    FilerReset;
    exit;
  end;

  assignFile (WorkFile, strPas (LocalName));
  rewrite (WorkFile, 1);
  if IOResult <> 0 then begin
    Result := FS_FILE_WRITEERROR;
    FilerReset;
    exit;
  end;

  if not FilerRequest (CMD_INIT_GET) then begin
    Result := FS_FILE_NOTFOUND;
    FilerReset;
    exit;
  end;

  if ProgressProc (PluginNumber, RemoteName, LocalName, 0) <> 0 then begin
    Result := FS_FILE_USERABORT;
    closeFile (WorkFile);
    FilerReset;
    FilerRequest (CMD_DATA_END);
    exit;
  end;

  Transferred := 0;
  FileLength  := RemoteInfo^.SizeLow;
  new (PBuffer);

  try
// ------- main copy loop -------
    repeat
      Packet.size := PACKET_DATA_SIZE;
      Packet.data := PBuffer;
      if not FilerRequest (CMD_GET_DATA) then begin
        closeFile (WorkFile);
        FilerReset;
        FilerRequest (CMD_DATA_END);
        Result := FS_FILE_READERROR;
        exit;
      end;

      Delivered   := Packet.size;
      Transferred := Transferred + Delivered;
      if FileLength = 0 then
        PercentDone := 100
      else
        PercentDone := (Transferred * 100) div FileLength;

      BlockWrite (WorkFile, PBuffer^, Delivered);
      if IOResult <> 0 then begin
        closeFile (WorkFile);
        Result := FS_FILE_WRITEERROR;
        FilerReset;
        FilerRequest (CMD_DATA_END);
        exit;
      end;

      if ProgressProc (PluginNumber, RemoteName, LocalName, PercentDone) <> 0 then begin
        closeFile (WorkFile);
        Result := FS_FILE_USERABORT;
        FilerReset;
        FilerRequest (CMD_DATA_END);
        exit;
      end;

    until Delivered < PACKET_DATA_SIZE;
// ------- end of main copy loop -------    

    SetFileTime (TFileRec (WorkFile).Handle, nil, nil, @RemoteInfo^.LastWriteTime);
    closeFile (WorkFile);
    if IOResult <> 0 then begin
      Result := FS_FILE_WRITEERROR;
      FilerReset;
      exit;
    end;

    if not FilerRequest (CMD_DATA_END) then begin
      Result := FS_FILE_READERROR;
      FilerReset;
      exit;
    end;

    SetFileAttributes (LocalName, RemoteInfo^.Attr);
 
    if (CopyFlags and FS_COPYFLAGS_MOVE) <> 0 then
      if not FilerRequest (CMD_DEL_FILE, RemoteName) then begin
        Result := FS_FILE_READERROR;
        FilerReset;
        exit;
      end;

    Result := FS_FILE_OK;

  finally
    dispose (PBuffer);
  end;
end;                  // function FsGetFile

{ ------------------------------------------------------------------ }

function FsPutFile (LocalName, RemoteName : pchar; CopyFlags : integer) : integer; stdcall;

var
  FileLength, Transferred, Delivered : longword;
  PercentDone : integer;
  WorkFile    : file;
  PBuffer     : TPData;
  FName       : CharArr;
  dirEntry    : TDirEntry;
  SRemoteName : string;

begin
  SRemoteName := RemoteName + 1;
  SRemoteName := extractFilePath (SRemoteName) + trimFileName (extractFileName (SRemoteName));
  strCopy (FName, pchar (SRemoteName));

// does remote file exist?
  if (CopyFlags and FS_COPYFLAGS_OVERWRITE) = 0 then begin
    if not getDirEntry (FName, DirEntry) then begin
      Result := FS_FILE_WRITEERROR;
      exit;
    end;
    if not Packet.endOfDir then begin
      Result := FS_FILE_EXISTS;
      exit
    end;  
  end
  else   // if exists, delete it before copying 
    if not FilerRequest (CMD_DEL_FILE, FName) then begin
      Result := FS_FILE_WRITEERROR;
      FilerReset;
      exit;
    end;

  LogProc (PluginNumber, MsgTYPE_DETAILS, PChar ('Copy ' + LocalName + ' to ' + FName));

  assignFile (WorkFile, strpas (LocalName));
  FileMode := 0;   // open file only for reading
  reset (WorkFile, 1);
  if IOResult <> 0 then begin
    Result := FS_FILE_READERROR;
    FilerReset;
    exit;
  end;

  if not FilerRequest (CMD_SEND_FILENAME, FName) then begin
    Result := FS_FILE_NOTFOUND;
    FilerReset;
    exit;
  end;

  if ProgressProc (PluginNumber, @FName, LocalName, 0) <> 0 then begin
    Result := FS_FILE_USERABORT;
    closeFile (WorkFile);
    FilerRequest (CMD_DATA_END);
    if not FilerRequest (CMD_DEL_FILE, FName) then
      FilerReset;
    exit;
  end;

  Transferred := 0;
  FileLength  := fileSize (WorkFile);
  new (PBuffer);

  try
// --------- main copy loop -----------  
    repeat
      BlockRead (WorkFile, PBuffer^, PACKET_DATA_SIZE, Delivered);
      if IOResult <> 0 then begin
        closeFile (WorkFile);
        FilerRequest (CMD_DATA_END);
        if not FilerRequest (CMD_DEL_FILE, FName) then
          FilerReset;
        Result := FS_FILE_READERROR;
        exit;
      end;

      Packet.size := Delivered;
      Packet.data := PBuffer;
      if not FilerRequest (CMD_SEND_DATA) then begin
        closeFile (WorkFile);
        Result := FS_FILE_WRITEERROR;
        FilerReset;
        FilerRequest (CMD_DATA_END);
        exit;
      end;

      Transferred := Transferred + Delivered;
      if FileLength = 0 then
        PercentDone := 100
      else
        PercentDone := (Transferred * 100) div FileLength;

      if ProgressProc (PluginNumber, @FName, LocalName, PercentDone) <> 0 then begin
        closeFile (WorkFile);
        FilerRequest (CMD_DATA_END);
        Result := FS_FILE_USERABORT;
        FilerReset;
        exit;
      end;

    until Delivered < PACKET_DATA_SIZE;
// -------- end of main copy loop --------

    closeFile (WorkFile);
    if IOResult <> 0 then begin
      Result := FS_FILE_READERROR;
      FilerReset;
      exit;
    end;

    if not FilerRequest (CMD_DATA_END) then begin
      Result := FS_FILE_WRITEERROR;
      FilerReset;
      exit;
    end;

    if (CopyFlags and FS_COPYFLAGS_MOVE) <> 0 then
      if not deleteFile (LocalName) then begin
        Result := FS_FILE_READERROR;
        FilerReset;
        exit;
      end;

    Result := FS_FILE_OK;

  finally
    dispose (PBuffer);
  end;
end;       // function FsPutFile

{ ------------------------------------------------------------------ }

function FsRenMovFile (OldName, NewName : pchar; Move, OverWrite : bool;
         RemoteInfo : pRemoteInfo) : integer; stdcall;

begin
  inc (OldName);
  inc (NewName);

  { moving to different drive is not supported }
  if OldName[0] <> NewName[0] then begin
    Result := FS_FILE_NOTSUPPORTED;
    exit
  end;

  LogProc (PluginNumber, MsgTYPE_DETAILS, Pchar('Ren ' + OldName + ' to ' + NewName));
  
  Packet.size2 := strLen (NewName);
  Packet.data2 := TPData (NewName);
  if not FilerRequest (CMD_REN_FILE, OldName) then begin
    Result := FS_FILE_NOTFOUND;
    FilerReset;
    exit;
  end;

  Result := FS_FILE_OK;
end;
  
{ ------------------------------------------------------------------ }

function FsExecuteFile (MainWin : thandle; RemoteName, Verb : pchar) : integer; stdcall;

begin
  if Verb = 'open' then begin
    Result := FS_EXEC_YOURSELF;
    exit;
  end;
  
  if (Verb = 'properties') and (RemoteName = '\') then
    ShellExecute (0, nil, 'Notepad.exe', PChar (IniFileName), '', SW_SHOWNORMAL);
    
  Result := FS_EXEC_OK;
end;

{ ------------------------------------------------------------------ }

procedure FsGetDefRootName (DefRootName : pchar; maxlen : integer); stdcall;

begin
  strLCopy (DefRootName, PLUGIN_TITLE, maxlen - 1);
end;

{ ------------------------------------------------------------------ }

procedure FsSetDefaultParams (dps : pFsDefaultParamStruct); stdcall;

begin
  IniFileName := dps^.DefaultIniName;
end;

{ ------------------------------------------------------------------ }

exports
  FsDeleteFile,
  FsDisconnect,
  FsExecuteFile,
  FsFindClose,
  FsFindFirst,
  FsFindNext,
  FsGetDefRootName,
  FsGetFile,
  FsInit,
  FsMkDir,
  FsPutFile,
  FsRemoveDir,
  FsRenMovFile,
  FsSetDefaultParams;
  
{ ------------------------------------------------------------------ }

begin
  IniFileName := 'fsplugin.ini';
end.